#!/usr/bin/env python3
from spacemake.contrib import __author__, __email__, __version__
import pysam
import argparse
import os
import sys
import logging


def read_header(source):
    hd = {}
    for line in open(source):
        if not line.startswith("@"):
            continue

        parts = line[1:].split("\t")
        top = parts[0]
        rec = {}
        for entry in parts[1:]:
            key, value = entry.rstrip().split(":", maxsplit=1)
            rec[key] = value

        if top in ["SQ", "PG", "RG"]:
            if not top in hd:
                hd[top] = []
            hd[top].append(rec)
        else:
            hd[top] = rec

    return hd


def make_ID_unique(ids1, id2):
    id2_ = str(id2)
    n = 0
    while id2_ in ids1:
        # print(f"while enter: {id2_}")
        try:
            if "." in id2_:
                n = int(id2_.split(".")[-1])
                id2_ = id2_.rsplit(".", maxsplit=1)[0]
        except ValueError:
            pass

        n += 1
        id2_ = f"{id2_}.{n}"

    return id2_


def repair_PG_lines(pg_list, fix=False):
    from collections import defaultdict

    issues = []

    no_PP = []
    w_PP = []

    ids_seen = set()
    ids_by_cmd = defaultdict(list)
    rename = {}

    for i, pg in enumerate(pg_list):
        cmd = pg["ID"].rsplit(".", maxsplit=1)[0]
        if fix:
            pg["ID"] = rename.get(pg["ID"], pg["ID"])

        if "PP" in pg:
            if fix:
                pg["PP"] = rename.get(pg["PP"], pg["PP"])

            w_PP.append(pg)
            if pg["ID"] == pg["PP"]:
                best_guess = ids_by_cmd[cmd][-1]
                issues.append(
                    f"entry ID={pg['ID']} references itself: PP={pg['PP']}. Best guess: correct PP={best_guess}"
                )
                if fix:
                    # correct entry is most likely previous run of same tools
                    pg["PP"] = best_guess

            if pg["PP"] not in ids_seen:
                issues.append(
                    f"entry ID={pg['ID']} references PP={pg['PP']}, which has not been defined yet. At a minimum, the ordering is off."
                )
        else:
            no_PP.append(pg)
            if i > 0:
                best_guess = pg_list[i - 1]["ID"]
                issues.append(
                    f"subsequent PG entry without PP tag ID={pg['ID']}. Have to guess that {no_PP[0]['ID']} is actual beginning. Best guess: correct PP={best_guess}"
                )

                if fix:
                    pg["PP"] = best_guess

        if pg["ID"] in ids_seen:
            issues.append(f"multiple use of ID {pg['ID']}.")
            if fix:
                new_ID = make_ID_unique(ids_seen, pg["ID"])
                rename[pg["ID"]] = new_ID
                pg["ID"] = new_ID

        ids_seen.add(pg["ID"])
        ids_by_cmd[cmd].append(pg["ID"])

    if fix:
        pg_list = sort_PG_list(pg_list)

    return issues, pg_list


def header_to_SAM(header):
    SAM = []
    for k, v in header.items():
        if type(v) == dict:
            vstr = "\t".join([f"{x}:{y}" for x, y in v.items()])
            SAM.append(f"@{k}\t{vstr}")
        elif type(v) == list:
            for row in v:
                if type(row) == dict:
                    keys = list(row.keys())
                    if k == "PG":
                        # override order to ensure CL is always last for max compatibility
                        keys = [x for x in ["ID", "PN", "VN", "PP", "CL"] if x in row]

                    vstr = "\t".join([f"{x}:{row[x]}" for x in keys])
                else:
                    vstr = str(row)
                SAM.append(f"@{k}\t{vstr}")
        else:
            SAM.append(f"@{k}\t{v}")

    return SAM


def print_header(header):
    print("\n".join(header_to_SAM(header)))


def sort_PG_list(pg_list):
    root = None
    pp = {}
    for pg in pg_list:
        if not "PP" in pg:
            # no Previous Program: we have the first entry
            root = pg
        else:
            pp[pg["PP"]] = pg

    l = [root]
    pp_id = root["ID"]
    for i in range(len(pg_list) - 1):
        pg = pp[pp_id]
        l.append(pg)
        pp_id = pg["ID"]

    return l


def unique_IDs(pg_list1, pg_list2):
    from collections import defaultdict

    id_renames = {}
    ids1 = set([pg["ID"] for pg in pg_list1])
    # print(f"pre-existing IDs: {ids1}")
    for pg in pg_list2:
        id2 = pg["ID"]
        id2_ = str(id2)
        n = 0
        while id2_ in ids1:
            # print(f"while enter: {id2_}")
            try:
                if "." in id2_:
                    n = int(id2_.split(".")[-1])
                    id2_ = id2_.rsplit(".", maxsplit=1)[0]
            except ValueError:
                pass

            n += 1
            id2_ = f"{id2_}.{n}"

        # print(f"end result: {id2} -> {id2_}")
        if id2_ != id2:
            ids1.add(id2_)
            id_renames[id2] = id2_

    for pg in pg_list2:
        pg["ID"] = id_renames.get(pg["ID"], pg["ID"])
        if "PP" in pg:
            pg["PP"] = id_renames.get(pg["PP"], pg["PP"])

    # print("suggest the following renames in pg_list2:")
    # print(id_renames)

    return pg_list2


test_pg_list1 = [
    {"ID": "fastq_to_uBAM"},
    {"ID": "bowtie2", "PP": "fastq_to_uBAM"},
]

test_pg_list2 = [
    {"ID": "bowtie2"},
]


def merge_PG_lists(orig, other):
    first = sort_PG_list(orig.copy())
    last = sort_PG_list(other.copy())

    # print("about to merge PG lists")
    # for pg in first:
    #     print("first:", pg)

    # for pg in last:
    #     print("last:", pg)

    last = unique_IDs(
        first, last
    )  # rename re-occuring IDs to make them unique if needed
    last[0]["PP"] = first[-1]["ID"]  # link the two histories
    return first + last


def fix_PP(repair_header, fix=False):
    """
    Due to a bug in previous versions, we may end up with ID:bowtie2.2 and PP:bowtie2.2
    in the same PG entry. This breaks samtools view (unless --no-PG is used). So, here we
    detect this very specific error and fix it.
    """
    import pysam

    bam = pysam.AlignmentFile(repair_header, "rc", check_sq=False)
    hdr = bam.header.to_dict()
    issues, pg_list = repair_PG_lines(hdr["PG"], fix=True)
    for iss in issues:
        print(f"Header issue detected: '{iss}'.")

    if len(issues) == 0:
        print("No issues detected.")
    else:
        issues_after_fix, pg_list = repair_PG_lines(hdr["PG"], fix=False)
        assert len(issues_after_fix) == 0
        print(f"Successfully generated a fixed header.")

        from tempfile import NamedTemporaryFile

        if fix:
            with NamedTemporaryFile(mode="tw") as tmp_hdr:
                tmp_hdr.write("\n".join(header_to_SAM(hdr)) + "\n")
                tmp_hdr.flush()
                # tmp_hdr.close()
                import pysam

                pysam.samtools.reheader("-P", "-i", tmp_hdr.name, args.repair_header)
                print(
                    f"Replaced original header of '{args.repair_header}' with fixed header."
                )


def merge_headers(orig, other, enforce_RG=True):
    merged = dict(orig)
    # start from the original, including VN and RG entries...
    # connect the processing chains:
    # most recent program should be on top
    #  Previous Program (PP) of the first new output was the last Program Name (PN) in the original uBAM
    # other["PG"][-1]["PP"] = orig["PG"][0]["ID"]
    merged["PG"] = merge_PG_lists(
        orig["PG"], other["PG"]
    )  # unique_IDs(merged["PG"] + other["PG"])

    if "SO" in other["HD"]:
        merged["HD"]["SO"] = other["HD"]["SO"]  # keep sort-order

    # sequence identifiers should be absent from uBAM and at any rate are overwritten here
    merged["SQ"] = other["SQ"]
    if enforce_RG and not "RG" in merged or len(merged["RG"]) == 0:
        merged["RG"] = {"ID": "A", "SM": "NA"}
    # merged['HD']['SO'] = star['HD']['SO']  # sorted by

    return merged


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=(
            "STAR and bowtie2 create a new header from scratch and ignore everything upstream. "
            "This script fixes the .bam headers of such mapped output by splicing it together with "
            "the original uBAM header."
        )
    )

    parser.add_argument(
        "--in-bam",
        help="mapped star/bowtie2 bam input (default=/dev/stdin)",
        default="/dev/stdin",
    )
    parser.add_argument("--in-ubam", help="unmapped dropseq tagged bam")
    parser.add_argument(
        "--out-bam",
        help="fixed output bam (default=/dev/stdout)",
        default="/dev/stdout",
    )
    parser.add_argument("--out-mode", help="mode for output (default=b0)", default="b0")
    parser.add_argument("--in-header1", help="first header")
    parser.add_argument("--in-header2", help="second header")
    parser.add_argument("--repair-header", default="")
    parser.add_argument("--check-header", default="")

    args = parser.parse_args()

    header1 = None
    header2 = None

    if args.repair_header:
        fix_PP(args.repair_header, fix=True)
    elif args.check_header:
        fix_PP(args.check_header, fix=False)
    else:
        if args.in_bam and args.in_ubam:
            mbam = pysam.AlignmentFile(args.in_bam, "r")
            ubam = pysam.AlignmentFile(args.in_ubam, "rc", check_sq=False)
            header1 = ubam.header.to_dict()
            header2 = mbam.header.to_dict()

        if args.in_header1 and args.in_header2:
            # overwrite headers if so desired
            header1 = read_header(args.in_header1)
            header2 = read_header(args.in_header2)

        # print_header(header1)
        # print_header(header2)
        if header1 and header2:
            merged_header = merge_headers(header1, header2)
        # print_header(merged_header)
        # print(f"mapped BAM header")
        # print_header(header1)

        # print(f"original uBAM header")
        # print_header(ubam_header)

        # print("merged header")
        # print_header(merged_header)

        # copy input to output, just with the new header
        if args.in_bam and header1 and header2:
            bam_out = pysam.AlignmentFile(
                args.out_bam, f"w{args.out_mode}", header=merged_header
            )
            for aln in mbam.fetch(until_eof=True):
                bam_out.write(aln)



import base64
import datetime
import io
import logging
import os
from itertools import cycle
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from jinja2 import Template
from scipy.stats import gaussian_kde
from spacemake.config import ConfigFile
from spacemake.project_df import ProjectDF
from spacemake.util import message_aggregation

absolute_path = os.path.dirname(__file__)

cpalette = {
    "grey": "#999999",
    "light_orange": "#E69F00",
    "light_blue": "#56B4E9",
    "green": "#009E73",
    "yellow": "#F0E442",
    "blue": "#0072B2",
    "orange": "#D55E00",
    "pink": "#CC79A7",
}

clrs = {
    "umis": cpalette["light_orange"],
    "genes": cpalette["light_blue"],
    "reads": cpalette["green"],
    "pcr": cpalette["pink"],
    "pct_counts_mt": "black",
}

SAMPLEINFO_VARS = [
    "species",
    "sequencing_date",
    "investigator",
    "experiment",
    "barcode_flavor",
    "sequencing_date",
    "puck",
]

parula_dict = {
    1: "#352a87",
    2: "#2058b0",
    3: "#1f7eb8",
    4: "#28a7d7",
    5: "#38d7e3",
    6: "#99d4d0",
    7: "#aacca1",
    8: "#bbcc74",
    9: "#cbcc49",
    10: "#e0d317",
}

PCT_DOWNSAMPLE_TO_PLOT = [20, 40, 60, 80, 100]
DOWNSAMPLE_PCTS = list(range(10, 110, 10))

logger_name = "spacemake.report.saturation_analysis"
logger = logging.getLogger(logger_name)


def setup_parser(parser):
    """
    Set up command-line arguments for the script.

    :param parser: Argument parser object.
    :type parser: argparse.ArgumentParser
    :returns: Updated argument parser object.
    :rtype: argparse.ArgumentParser
    """
    # These allow to get the run_mode_variables.* from the config file
    # and pbf_metrics.px_by_um via project_df.puck_barcode_file_metrics
    parser.add_argument(
        "--project-id",
        type=str,
        help="the project_id in spacemake project_df.csv",
        required=True,
    )
    parser.add_argument(
        "--sample-id",
        type=str,
        help="the sample_id in spacemake project_df.csv",
        required=True,
    )
    parser.add_argument(
        "--run-modes",
        type=str,
        nargs="+",
        help="run modes of the sample for which the report will be generated",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--downsampled-dge-summary",
        type=str,
        nargs="+",
        help="path to the 'downsampled_dge_summary' file(s) generated by spacemake downsampling analysis.",
        required=True,
    )
    parser.add_argument(
        "--puck-barcode-file-id",
        type=str,
        help="the puck_barcode_file_id for the current report",
        required=True,
    )
    # This specifies where the output file will be generated
    parser.add_argument(
        "--out-html-report",
        type=str,
        help="where the HTML report will be saved",
        required=True,
    )

    return parser


def plot_density_metric_faceted(values, metric, log_scale=True, color="#000000", title=""):
    fig, axes = plt.subplots(len(PCT_DOWNSAMPLE_TO_PLOT), 1, figsize=(5, 0.5 * len(PCT_DOWNSAMPLE_TO_PLOT)))

    i = 0
    for downsample_pct, value_density in values.groupby("_downsample_pct_report"):
        if int(downsample_pct) in PCT_DOWNSAMPLE_TO_PLOT:
            density_function = gaussian_kde(np.nan_to_num(value_density[metric]), bw_method=0.1)
            x = np.linspace(1, max(np.nan_to_num(values[metric])), 100)

            axes[i].plot(x, density_function(x), color="black", linewidth=1)
            axes[i].fill_between(x, density_function(x), color=color)
            axes[i].set_yticks([])

            if log_scale:
                axes[i].set_xscale("log")

            axes[i].spines[["right", "top", "bottom"]].set_visible(False)
            axes[i].text(1.05, 0.5, f"{downsample_pct}%", transform=axes[i].transAxes, va="center")
            i += 1

        axes[-1].spines[["right", "top"]].set_visible(False)
        axes[-1].spines[["left", "bottom"]].set_visible(True)
        axes[-1].set_xlabel(title)

    for i in range(i - 1):
        axes[i].set_xticks([])

    fig.text(0.0, 0.6, "density", va="center", rotation="vertical")
    plt.tight_layout()

    return fig, axes


def plot_median_per_run_mode(values, metric, umi_cutoffs, color="#000000", title=""):
    fig, axes = plt.subplots(1, 1, figsize=(5, 3))

    lines = ["-", "--", "-.", ":"]
    linecycler = cycle(lines)
    handles, labels = [], []

    for umi_cutoff in umi_cutoffs:
        _values = values[values["total_counts"] > umi_cutoff]
        median_values = (
            _values[[metric, "_downsample_pct_report"]].groupby("_downsample_pct_report").median().reset_index()
        )

        linestyle = next(linecycler)

        (line,) = axes.plot(
            median_values["_downsample_pct_report"], median_values[metric], linestyle, color=color, label=umi_cutoff
        )
        axes.scatter(
            median_values["_downsample_pct_report"], median_values[metric], s=20, color=color, edgecolors="black"
        )

        handles.append(line)
        labels.append(umi_cutoff)

    axes.set_xticks(PCT_DOWNSAMPLE_TO_PLOT)
    axes.set_xticklabels([f"{pct}%" for pct in PCT_DOWNSAMPLE_TO_PLOT])
    axes.spines[["right", "top"]].set_visible(False)
    axes.set_xlabel("downsampling percentage")
    axes.set_ylabel(title)

    legend = axes.legend(handles, labels, loc="lower right", title="UMI cutoff")
    legend.set_frame_on(False)

    plt.tight_layout()
    return fig, axes


def generate_deciled_data(values):
    # Group by '_downsample_pct_report' and perform the necessary calculations
    def calculate_deciles(group):
        group["cumsum_reads"] = group["n_reads"].cumsum()
        group["decile_limit"] = group["n_reads"].sum() / 10
        group["decile"] = (group["cumsum_reads"] / group["decile_limit"]).floordiv(1) + 1
        return group.loc[group["decile"] < 11]

    # Group by '_downsample_pct_report' and apply the calculate_deciles function
    decile_dat = values.groupby("_downsample_pct_report").apply(calculate_deciles).reset_index(drop=True)

    # Group by 'percentage' and 'decile' and calculate medians and counts
    decile_dat = (
        decile_dat.groupby(["_downsample_pct_report", "decile"])
        .agg(
            {
                "n_reads": "median",
                "n_genes_by_counts": "median",
                "reads_per_counts": "median",
                "total_counts": "median",
                "cell_bc": "count",
            }
        )
        .reset_index()
    )

    # Melt the DataFrame to long format
    decile_dat = pd.melt(
        decile_dat, id_vars=["_downsample_pct_report", "decile"], var_name="observation", value_name="value"
    )

    # Convert 'decile' and '_downsample_pct_report' to appropriate data types
    decile_dat["decile"] = decile_dat["decile"].astype("category")
    decile_dat["_downsample_pct_report"] = decile_dat["_downsample_pct_report"].astype(int)

    mapping_dict = {
        "n_reads": "median_reads",
        "n_genes_by_counts": "median_genes",
        "reads_per_counts": "median_pcr",
        "total_counts": "median_umis",
        "cell_bc": "n_beads",
    }
    decile_dat["observation"] = decile_dat["observation"].replace(mapping_dict)

    return decile_dat


def plot_deciled_median(decile_dat):
    fig, axes = plt.subplots(3, 2, figsize=(6, 4))

    # Iterate through each unique 'observation' for facetting
    for i, (obs, data) in enumerate(decile_dat.groupby("observation")):
        for _obs, _data in data.groupby("decile"):
            axes.flatten()[i].plot(
                _data["_downsample_pct_report"], _data["value"], label=_obs, linewidth=0.6, color=parula_dict[_obs]
            )
            axes.flatten()[i].scatter(
                _data["_downsample_pct_report"], _data["value"], s=20, edgecolors="black", color=parula_dict[_obs]
            )

        axes.flatten()[i].set_xticks([0, 20, 40, 60, 80, 100])
        axes.flatten()[i].set_xticklabels(["0", "20", "40", "60", "80", "100"])
        axes.flatten()[i].set_title(obs)
        axes.flatten()[i].spines[["top", "right"]].set_visible(False)

    axes.flatten()[i].set_xlabel("downsampling percentage")
    if i % 2 == 0:
        axes.flatten()[-1].axis("off")

    # Create a single legend at the bottom
    handles, labels = [], []
    for obs in parula_dict:
        handles.append(plt.Line2D([0], [0], marker="o", color="w", markerfacecolor=parula_dict[obs], markersize=8))
        labels.append(str(obs))

    fig.legend(handles, labels, title="Decile", loc="lower right", ncol=3, bbox_to_anchor=(0.95, 0.02))

    plt.tight_layout()

    return fig, axes


def plot_to_base64(fig):
    my_stringIObytes = io.BytesIO()
    fig.savefig(my_stringIObytes, format="jpg", dpi=300)
    plt.close()
    my_stringIObytes.seek(0)
    return base64.b64encode(my_stringIObytes.read()).decode()


def load_dge_summary_downsampling(downsampled_dge_summary, run_mode, puck_barcode_file_id):
    obs_df = pd.DataFrame()
    for downsample_pct in DOWNSAMPLE_PCTS:
        _obs_df = pd.read_csv(
            downsampled_dge_summary[f"downsampled_dge_summary.{run_mode}.{downsample_pct}.{puck_barcode_file_id}"]
        )
        _obs_df["_downsample_pct_report"] = downsample_pct
        obs_df = pd.concat([obs_df, _obs_df])

    return obs_df


def generate_saturation_analysis_metadata(
    project_id: str,
    sample_id: str,
    run_modes: Union[list, str],
    downsampled_dge_summary: Union[dict, list, str],
    puck_barcode_file_id: str,
):
    if isinstance(downsampled_dge_summary, str):
        downsampled_dge_summary = [downsampled_dge_summary]
    if isinstance(run_modes, str):
        run_modes = [run_modes]
    if (
        isinstance(run_modes, list)
        and isinstance(downsampled_dge_summary, list)
        and len(run_modes) == len(downsampled_dge_summary)
    ):
        downsampled_dge_summary = {
            f"downsampled_dge_summary.{run_mode}.{d_pct}.{puck_barcode_file_id}": d
            for run_mode, d, d_pct in zip(run_modes, downsampled_dge_summary, DOWNSAMPLE_PCTS)
        }
    elif (
        isinstance(run_modes, list)
        and isinstance(downsampled_dge_summary, list)
        and len(run_modes) != len(downsampled_dge_summary)
    ):
        raise ValueError("'run_modes' and 'downsampled_dge_summary' must have the same length")

    report = {
        "type": "saturation_analysis",
        "runinformation": [],
        "date": datetime.date.today().strftime("%Y/%m/%d"),
        "plots": [],
    }
    main_plots = {"histostats": [], "medianstats": [], "deciledmedian": []}
    report["plots"] = main_plots

    # Loading project_df for metadata
    config = ConfigFile.from_yaml("config.yaml")
    project_df = ProjectDF("project_df.csv", config=config)

    # Table: sample info
    sample_info = project_df.get_sample_info(project_id, sample_id)
    report["sampleinfo"] = {}
    report["sampleinfo"]["project_id"] = project_id
    report["sampleinfo"]["sample_id"] = sample_id
    report["sampleinfo"].update({svar: sample_info[svar] for svar in SAMPLEINFO_VARS})

    # Loading all dge data summaries
    # Table: summarised metrics over beads
    dge_summaries = {}
    for run_mode in run_modes:
        dge_summaries[run_mode] = {}
        dge_summaries[run_mode] = load_dge_summary_downsampling(
            downsampled_dge_summary, run_mode, puck_barcode_file_id
        )

    # Plots
    # Histograms per run mode
    for run_mode, dge_summary in dge_summaries.items():
        umicutoff = {"name": run_mode, "umiplot": None, "readsplot": None, "readsumiplot": None}
        fig, _ = plot_density_metric_faceted(
            dge_summary, "total_counts", log_scale=True, color=clrs["umis"], title="# of UMIs per spatial unit"
        )
        umicutoff["umiplot"] = plot_to_base64(fig)

        fig, _ = plot_density_metric_faceted(
            dge_summary, "n_reads", log_scale=True, color=clrs["reads"], title="# of reads per spatial unit"
        )
        umicutoff["readsplot"] = plot_to_base64(fig)

        fig, _ = plot_density_metric_faceted(
            dge_summary, "reads_per_counts", log_scale=False, color=clrs["pcr"], title="reads/UMI per spatial unit"
        )
        umicutoff["readsumiplot"] = plot_to_base64(fig)

        report["plots"]["histostats"].append(umicutoff)

    # Median plots per run mode
    for run_mode, dge_summary in dge_summaries.items():
        umi_cutoffs_values = project_df.config.get_run_mode(run_mode).variables["umi_cutoff"]
        umi_cutoffs_values = list(sorted(list(set([int(u) for u in umi_cutoffs_values] + [1]))))
        medianstats = {"name": run_mode, "umiplot": None, "readsplot": None, "readsumiplot": None}
        fig, _ = plot_median_per_run_mode(
            dge_summary, "total_counts", umi_cutoffs_values, color=clrs["umis"], title="median reads\nper spatial unit"
        )
        medianstats["umiplot"] = plot_to_base64(fig)

        fig, _ = plot_median_per_run_mode(
            dge_summary, "n_reads", umi_cutoffs_values, color=clrs["reads"], title="median UMIs\nper spatial unit"
        )
        medianstats["readsplot"] = plot_to_base64(fig)

        fig, _ = plot_median_per_run_mode(
            dge_summary,
            "reads_per_counts",
            umi_cutoffs_values,
            color=clrs["pcr"],
            title="median reads/UMI\nper spatial unit",
        )
        medianstats["readsumiplot"] = plot_to_base64(fig)

        report["plots"]["medianstats"].append(medianstats)

    # Deciled plots
    for run_mode, dge_summary in dge_summaries.items():
        deciledmedian = {"name": run_mode, "plot": None}
        decile_dat = generate_deciled_data(dge_summary)
        fig, _ = plot_deciled_median(decile_dat)
        deciledmedian["plot"] = plot_to_base64(fig)
        report["plots"]["deciledmedian"].append(deciledmedian)

    return report


def generate_html_report(data, template_file):
    with open(template_file, "r") as template_data:
        template_content = template_data.read()
        template = Template(template_content)

    html_report = template.render(report=data)

    return html_report


@message_aggregation(logger_name)
def cmdline():
    """cmdline."""
    import argparse

    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="generate spacemake's 'saturation analysis' with python",
    )
    parser = setup_parser(parser)

    args = parser.parse_args()
    template_file = os.path.join(absolute_path, "templates/saturation_analysis.html")

    report_metadata = generate_saturation_analysis_metadata(
        args.project_id, args.sample_id, args.run_modes, args.downsampled_dge_summary, args.puck_barcode_file_id
    )

    html_report = generate_html_report(report_metadata, template_file)

    with open(args.out_html_report, "w") as output:
        output.write(html_report)


if __name__ == "__main__":
    cmdline()



import pysam
import pandas as pd
from spacemake.util import FASTQ_src, read_fq
from byo.track import load_track

genome = load_track("/data/rajewsky/genomes/hg38/hg38.fa")

# load output of gene_loci_to_gtf.py: a table with gene-name, start and end coordinates
df_genes = pd.read_csv(
    "chr22_gene_bounds.csv", sep="\t", names=["gene", "start", "end", "L"]
).set_index("gene")

# keep track of the few dozen reads that Nikos has selected
selected_reads = {}
for fa_id, seq, qual in read_fq("reads_chr22_R2.fastq.gz"):
    if "IGLC3" in fa_id:
        print(f"selecting IGLC3 read {fa_id} -> {fa_id.split('_')[0]}")
    selected_reads[fa_id.split("_")[0]] = fa_id

# then go through the SAM file (no header) to get the mapping position of these reads (not ideal)
df = pd.read_csv(
    "/data/rajewsky/home/nkarais/murphy/fc_sts/collect_reads_chr22/final.polyA_adapter_trimmed_chr22.sam",
    sep="\t",
    header=None,
)

starts = []
for row in df.itertuples():
    # print(row)
    qname = row[1]
    if "A00643:496:HFJ5MDRX2:1:2101:12888:1172" in qname:
        print(f"YAY! detected IGLC3 read {qname}")

    if qname in selected_reads:
        print(f"selecting read {qname}")
        starts.append(row[4])

# find genes that overlap the selected reads mapping position
# this intersection code is very crude, but effective
intervals = set()
starts = set(starts)
for row in df_genes.itertuples():
    next_starts = set(starts)
    for x in starts:
        if row.start < x and row.end > x:
            intervals.add((row.start, row.end))
            print(f"selecting gene entry '{row}'")
            next_starts.discard(x)

    starts = next_starts

print(
    f"we have the following start coordinates left. selecting buffer regions around these"
)
print(starts)


def do_merge(s, e, intervals):
    keep = []
    for j, (s2, e2) in enumerate(intervals):
        new = (s2, e2)
        if s2 <= e and e2 >= e:
            print(f"overlap on the right, s={s} e={e} s2={s2} e2={e2}")
            new = (min(s2, s), max(e, e2))
        elif e2 >= s and s2 <= s:
            print(f"overlap on the left, s={s} e={e} s2={s2} e2={e2}")
            new = (min(s2, s), max(e, e2))
        elif s2 >= s and e2 <= e:
            print(f"contained in other interval. discard, s={s} e={e} s2={s2} e2={e2}")
            continue

        keep.append(new)

    return keep


# merge intervals that have some overlap
intervals = sorted(list(intervals), key=lambda x: x[1] - x[0], reverse=True)
print(intervals)

while True:
    changed = False
    for i, (s, e) in enumerate(intervals):
        others = intervals[i + 1 :]
        keep = do_merge(s, e, others)
        if keep != others:
            print("we had a change!")
            print("before")
            for s, e in intervals:
                print(f"{s} - {e}")

            intervals = intervals[: i + 1] + keep
            intervals = sorted(list(intervals), key=lambda x: x[1] - x[0], reverse=True)
            print("after")
            for s, e in intervals:
                print(f"{s} - {e}")

            changed = True
            break  # the for loop

    if not changed:
        break

intervals = sorted(list(set(intervals)), key=lambda x: x[1] - x[0], reverse=True)
print("remaining intervals")
for s, e in intervals:
    print(f"{s} - {e}")


# Okay, now we know the gene loci which are needed to map the test reads!
intervals = list(sorted(intervals))
print(f"relevant intervals found: {len(intervals)}")

# extract the genomic sequence for the loci we need and save as a mini-"genome"
with open("test_genome.fa", "wt") as f:
    for start, end in intervals:
        seq = genome.get_oriented("chr22", start, end, "+")
        f.write(f">test_chr22.{start}-{end}\n{seq}\n")

# cut down the GTF annotation to only those parts that pertain to the genic regions we care about
with open("test_annotation.gtf", "wt") as f:
    for line in open("gencode.v38.chr22.gtf", "rt"):
        if line.startswith("#"):
            continue

        parts = line.rstrip().split("\t")
        chrom, source, rec, start, end = parts[:5]
        if chrom != "chr22":
            continue

        start = int(start)
        end = int(end)

        for s, e in intervals:
            if (
                (start <= s and end > s)  # overlap the start of interval
                or (start > s and end < e)  # internal to interval
                or (start < e and end > e)  # overlap the end of interval
                or (start < s and end > e)  # overlap the entire interval
            ):
                # the name of the pseudo-chromosome this is on (see excision of genomic sequence above)
                chrom = f"test_{chrom}.{s}-{e}"

                start = max(
                    0, start - s
                )  # translate start and end coordinates from whole chr22 to the gene region
                end = min(e - s, end - s)

                parts[0:5] = (chrom, "test_data", rec, str(start), str(end))
                f.write("\t".join(parts) + "\n")

# Done! We now have:
# * test_genome.fa.gz with the gene sequences
# * test_annotation.gtf.gz with the gene models (exon/intron, CDS/UTR etc.)
# * reads_chr22_R1.fastq.gz with test read barcodes mapping to a few CBs and UMIs
# * reads_chr22_R2.fastq.gz with test read cDNAs mapping to the genic regions we care about



import logging
import anndata

import numpy as np

from typing import List, Union
from spacemake.util import message_aggregation

logger_name = "spacemake.spatial.puck_collection"
logger = logging.getLogger(logger_name)
DEFAULT_REGEX_PUCK_ID = "(L[1-4][a-b]_tile_[1-2][0-7][0-9][0-9])"


def setup_parser(parser):
    """
    Set up command-line arguments for the script.

    :param parser: Argument parser object.
    :type parser: argparse.ArgumentParser
    :returns: Updated argument parser object.
    :rtype: argparse.ArgumentParser
    """
    parser.add_argument(
        "--pucks",
        type=str,
        nargs="+",
        help="path to spatial.h5ad AnnData file, one per puck",
        required=True,
    )

    parser.add_argument(
        "--puck-id",
        type=str,
        nargs="+",
        help="list of puck id for the input files, same order as pucks."
        + "Must be specified when the filenames do not contain a puck id that can be parsed with --puck-id-regex",
        default=None,
    )

    parser.add_argument(
        "--puck-coordinates",
        type=str,
        help="name of puck collection",
        required=True,
    )

    parser.add_argument(
        "--output",
        type=str,
        help="path to output.h5ad AnnData file",
        required=True,
    )

    parser.add_argument(
        "--puck-id-regex",
        type=str,
        help="regex to find puck id in file names",
        default=DEFAULT_REGEX_PUCK_ID,
    )

    parser.add_argument(
        "--puck-id-key",
        type=str,
        help="name of .obs variable where puck id are (will be) stored",
        default="puck_id",
    )

    parser.add_argument(
        "--merge-output",
        type=str,
        help='how to merge pucks, can be "same", "unique", "first", "only"',
        choices=["same", "unique", "first", "only"],
        default="same",
    )

    parser.add_argument(
        "--join-output",
        type=str,
        help='how to join pucks, can be "inner", "outer"',
        choices=["inner", "outer"],
        default="outer",
    )

    parser.add_argument(
        "--no-reset-index",
        default=False,
        action="store_true",
        help="do not reset the obs_name index of the AnnData object as  'obs_name:<puck_id_key>'; keep original 'obs_name'",
    )

    parser.add_argument(
        "--no-transform",
        default=False,
        action="store_true",
        help="do not transform the spatial coordinates of the AnnData object",
    )

    return parser


def check_obs_unique(puck: anndata.AnnData, obs_key: str = "puck_id") -> bool:
    """
    Check if the observation keys in the AnnData object are unique.

    :param puck: AnnData object.
    :type puck: anndata.AnnData
    :param obs_key: Observation key to check uniqueness, defaults to "puck_id".
    :type obs_key: str, optional
    :returns: True if the observation keys are unique, False otherwise.
    :rtype: bool
    """
    return puck.obs[obs_key].nunique() == 1


def _transform_puck(puck: anndata.AnnData, pucks_transform: dict) -> anndata.AnnData:
    """
    Transform the spatial coordinates of the AnnData object.

    :param puck: AnnData object.
    :type puck: anndata.AnnData
    :param pucks_transform: Dictionary containing puck transformations.
    :type pucks_transform: dict
    :returns: Transformed AnnData object.
    :rtype: anndata.AnnData
    """
    if not check_obs_unique(puck, "puck_id"):
        raise ValueError(f"puck_id exists in AnnData object but are not unique")

    _puck_id = np.unique(puck.obs["puck_id"])[0]
    x_ofs = pucks_transform["x_offset"][_puck_id]
    y_ofs = pucks_transform["y_offset"][_puck_id]
    puck.obsm["spatial"][:, 0] += x_ofs
    puck.obsm["spatial"][:, 1] += y_ofs

    return puck


def create_puck_collection(
    puck: anndata.AnnData,
    puck_transform: dict,
    reset_index: bool = True,
    transform: bool = True,
) -> anndata.AnnData:
    """
    Create a puck collection with transformed spatial coordinates.

    :param puck: AnnData object.
    :type puck: anndata.AnnData
    :param puck_transform: Dictionary containing puck transformations.
    :type puck_transform: dict
    :param reset_index: Flag to reset the observation index, defaults to True.
    :type reset_index: bool, optional
    :param transform: Flag to transform the spatial coordinates, defaults to True.
    :type transform: bool, optional
    :returns: Puck collection with transformed spatial coordinates.
    :rtype: anndata.AnnData
    """
    if reset_index:
        puck.obs_names = (
            puck.obs_names.astype(str) + ":" + puck.obs["puck_id"].astype(str)
        )

    if transform:
        puck = _transform_puck(puck, puck_transform)

    return puck


def parse_puck_id_from_path(f: str, puck_id_regex: str = DEFAULT_REGEX_PUCK_ID) -> str:
    """
    Parse the puck ID from the file path using the specified regex.

    :param f: File path.
    :type f: str
    :param puck_id_regex: Regular expression to find the puck ID, defaults to DEFAULT_REGEX_PUCK_ID.
    :type puck_id_regex: str, optional
    :returns: Puck ID extracted from the file path.
    :rtype: str
    """
    import os
    import re

    bname = os.path.basename(f)
    puck_id = re.findall(rf"{puck_id_regex}", bname)

    if len(puck_id) > 1:
        logger.warn(
            "Found more than one puck_id in the path. First one (index 0) will be used."
        )

    puck_id = puck_id[0]

    return puck_id


def read_pucks_to_list(
    f: Union[str, List[str]],
    puck_id: Union[int, List[int], None] = None,
    puck_id_regex: str = DEFAULT_REGEX_PUCK_ID,
    puck_id_key: str = "puck_id",
) -> List:
    """
    Read the pucks from file(s) and return a list of AnnData objects.

    :param f: File path(s).
    :type f: Union[str, List[str]]
    :param puck_id: Puck ID(s), defaults to None.
    :type puck_id: Union[int, List[int], None], optional
    :param puck_id_regex: Regular expression to find the puck ID in file names, defaults to DEFAULT_REGEX_PUCK_ID.
    :type puck_id_regex: str, optional
    :param puck_id_key: Name of the variable where puck IDs are stored, defaults to "puck_id".
    :type puck_id_key: str, optional
    :returns: List of AnnData objects representing the pucks.
    :rtype: List
    """

    import scanpy as sc

    if type(f) is str:
        f = [f]

    if puck_id is not None and type(puck_id) is str:
        puck_id = [puck_id]
    elif type(puck_id) is list and len(puck_id) != len(f):
        raise ValueError(
            f"Dimensions for f ({len(f)}) and puck_id ({len(puck_id)}) are not compatible"
        )

    pucks = []

    for i, f in enumerate(f):
        _f_obj = sc.read_h5ad(f)

        if "spatial" not in _f_obj.obsm.keys():
            raise ValueError(f"Could not find valid .obsm['spatial'] data in {f}")
        if puck_id_key not in _f_obj.obs.keys():
            if puck_id is None:
                _puck_id = parse_puck_id_from_path(f)
                if len(_puck_id) == 0:
                    raise ValueError(
                        f"Could not find a puck_id from the filename {f} with the regular expression {puck_id_regex}"
                    )
            else:
                _puck_id = puck_id[i]

            _f_obj.obs[puck_id_key] = _puck_id

        if puck_id_key != "puck_id":
            _f_obj.obs["puck_id"] = _puck_id

        if not check_obs_unique(_f_obj, "puck_id"):
            raise ValueError(
                f"puck_id exist in AnnData object but are not unique for the puck in file {f}"
            )

        pucks.append(_f_obj)

    return pucks


def parse_puck_coordinate_system_file(f: str) -> dict:
    """
    Parse the puck coordinate system file and return a dictionary.

    :param f: File path of the coordinate system file.
    :type f: str
    :returns: Dictionary representing the puck coordinate system.
    :rtype: dict
    """

    import pandas as pd

    cs = pd.read_csv(f, sep="[,|\t]", engine="python")

    cs = cs.set_index("puck_id")

    cs = cs.loc[~cs.index.duplicated(keep="first")]

    return cs.to_dict(orient="dict")


def merge_pucks_to_collection(
    pucks: List[str],
    puck_id: List[str],
    puck_coordinates: str,
    puck_id_regex: str = None,
    puck_id_key: str = "puck_id",
    no_reset_index: bool = False,
    no_transform: bool = False,
    merge_output: str = "same",
    join_output: str = "outer",
) -> anndata.AnnData:
    """
    Merge multiple pucks into a single puck collection.

    :param pucks: List of puck file paths.
    :type pucks: List[str]
    :param puck_id: List of puck IDs corresponding to the input files.
    :type puck_id: List[str]
    :param puck_coordinates: Name of the puck collection.
    :type puck_coordinates: str
    :param puck_id_regex: Regular expression to find the puck ID in file names, defaults to None.
    :type puck_id_regex: str, optional
    :param puck_id_key: Name of the variable where puck IDs are stored, defaults to "puck_id".
    :type puck_id_key: str, optional
    :param no_reset_index: Flag to not reset the observation index, defaults to False.
    :type no_reset_index: bool, optional
    :param no_transform: Flag to not transform the spatial coordinates, defaults to False.
    :type no_transform: bool, optional
    :param merge_output: How to merge pucks, can be "same", "unique", "first", or "only", defaults to "same".
    :type merge_output: str, optional
    :param join_output: How to join pucks, can be "inner" or "outer", defaults to "inner".
    :type join_output: str, optional
    :returns: Merged puck collection as an AnnData object.
    :rtype: anndata.AnnData
    """

    puck_transform = parse_puck_coordinate_system_file(puck_coordinates)

    pucks_list = read_pucks_to_list(pucks, puck_id, puck_id_regex, puck_id_key)

    puck_collection_list = []

    for puck in pucks_list:
        puck_collection_list += [
            create_puck_collection(
                puck,
                puck_transform,
                ~no_reset_index,
                ~no_transform,
            )
        ]

    puck_collection = anndata.concat(
        puck_collection_list, merge=merge_output, join=join_output
    )

    puck_collection.uns = {np.unique(puck.obs[puck_id_key])[0]: puck.uns for puck in puck_collection_list}
    
    # we select the first puck from the collection, we assume the variables will be very similar across
    _current_puck = puck_collection_list[0]

    if "mesh_variables" in _current_puck.uns:
        puck_collection.uns["mesh_variables"] = _current_puck.uns["mesh_variables"]

    if "puck_variables" in _current_puck.uns:
        puck_collection.uns["puck_variables"] = _current_puck.uns["puck_variables"]

    return puck_collection

@message_aggregation(logger_name)
def cmdline():
    """cmdline."""
    import argparse

    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="stitching/transforming pucks into a common global coordinate system",
    )
    parser = setup_parser(parser)

    args = parser.parse_args()
    puck_collection = merge_pucks_to_collection(
        pucks=args.pucks,
        puck_id=args.puck_id,
        puck_coordinates=args.puck_coordinates,
        puck_id_regex=args.puck_id_regex,
        puck_id_key=args.puck_id_key,
        no_reset_index=args.no_reset_index,
        no_transform=args.no_transform,
        merge_output=args.merge_output,
        join_output=args.join_output,
    )

    puck_collection.write_h5ad(args.output)


if __name__ == "__main__":
    cmdline()



import errno
import os
import logging

# from contextlib import ContextDecorator, contextmanager
from spacemake.errors import SpacemakeError, FileWrongExtensionError
from spacemake.contrib import __version__, __license__, __author__, __email__

LINE_SEPARATOR = "-" * 50 + "\n"

bool_in_str = ["True", "true", "False", "false"]


def generate_kmers(k, nts="ACGT"):
    if k == 0:
        yield ""
    elif k > 0:
        for x in nts:
            for mer in generate_kmers(k - 1, nts=nts):
                yield x + mer


class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __str__(self):
        buf = ["dotdict"]
        for k, v in self.items():
            if not k.startswith("__"):
                buf.append(f"  {k} = {v}")

        return "\n".join(buf)


def wc_fill(x, wc):
    """
    Versatile snakemake helper function that can render any string template used in the mapping and reports modules,
    either filling in from a Wildcards object, or from a dotdict.
    """
    return x.format(
        sample_id=getattr(wc, "sample_id", "NO_sample_id"),
        project_id=getattr(wc, "project_id", "NO_project_id"),
        species=getattr(wc, "species", "NO_species"),
        ref_name=getattr(wc, "ref_name", "NO_ref_name"),
        mapper=getattr(wc, "mapper", "NO_mapper"),
        link_name=getattr(wc, "link_name", "NO_link_name"),
        polyA_adapter_trimmed=getattr(wc, "polyA_adapter_trimmed", ""),
        data_root_type=getattr(wc, "data_root_type", "complete_data"),
    )


def quiet_bam_open(*argc, **kw):
    """_summary_

    This wrapper around pysam.AlignmentFile() simply silences warnings about missing BAM index etc.
    We don't care about the index and therefore these error messages are just spam.

    Returns:
        pysam.AlignmentFile: the sam/bam object as returned by pysam.
    """
    import pysam

    save = pysam.set_verbosity(0)
    bam = pysam.AlignmentFile(*argc, **kw)
    pysam.set_verbosity(save)
    return bam


def assert_file(file_path, default_value=None, extension=["all"]):
    if file_path == default_value:
        # file doesn't exist but has the default value,
        # so we do not need to assert anything
        return False

    if isinstance(extension, str):
        extension = [extension]

    if not isinstance(file_path, list):
        file_path = [file_path]

    for fp in file_path:
        # check if file exists, raise error if not
        if not os.path.isfile(fp):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fp)

        # check for all extensions
        if extension != ["all"]:
            extension_check = [fp.endswith(ex) for ex in extension]
            if not any(extension_check):
                raise FileWrongExtensionError(fp, extension)

    # return true if file exists and every test was good
    return True


def str2bool(var):
    if isinstance(var, bool):
        return var

    if var in ["True", "true"]:
        return True
    elif var in ["False", "false"]:
        return False
    else:
        raise ValueError(f"variable should be boolean, or one of: {bool_in_str}")


def ensure_path(path):
    dirname = os.path.dirname(path)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    return path


def timed_loop(
    src,
    logger,
    T=5,
    chunk_size=10000,
    template="processed {i} BAM records in {dT:.1f}sec. ({rate:.3f} k rec/sec)",
    skim=0,
):
    from time import time

    t0 = time()
    t_last = t0
    i = 0
    for i, x in enumerate(src):
        if skim:
            if i % skim == 0:
                yield x
        else:
            yield x

        if i % chunk_size == 0:
            t = time()
            if t - t_last > T:
                dT = t - t0
                rate = i / dT / 1000.0
                logger.info(template.format(**locals()))
                t_last = t

    t = time()
    dT = t - t0
    rate = i / dT / 1000.0
    logger.info("Finished! " + template.format(**locals()))


def FASTQ_src(src):
    from more_itertools import grouper

    for name, seq, _, qual in grouper(src, 4):
        yield name.rstrip()[1:], seq.rstrip(), qual.rstrip()


def BAM_src(src):
    import pysam

    bam = quiet_bam_open(src, "rb", check_sq=False)
    for read in bam.fetch(until_eof=True):
        yield read.query_name, read.query_sequence, read.query_qualities


def read_fq(fname, skim=0):
    import gzip

    logger = logging.getLogger("spacemake.util.read_fq")
    if str(fname) == "None":
        logger.warning("yielding empty data forever")
        while True:
            yield ("no_qname", "no_seq", "no_qual")

    if "*" in fname:
        logger.warning("EXPERIMENTAL: fname contains wildcards")
        from glob import glob

        for match in sorted(glob(fname)):
            for rec in read_fq(match, skim=skim):
                yield rec

    logger.info(f"iterating over reads from '{fname}'")
    if fname.endswith(".gz"):
        src = FASTQ_src(gzip.open(fname, mode="rt"))
    elif fname.endswith(".bam"):
        src = BAM_src(fname)
    elif type(fname) is str:
        src = FASTQ_src(open(fname))
    else:
        src = FASTQ_src(fname)  # assume its a stream or file-like object already

    n = 0
    for record in timed_loop(
        src,
        logger,
        T=15,
        template="processed {i} reads in {dT:.1f}sec. ({rate:.3f} k rec/sec)",
        skim=skim,
    ):
        yield record
        n += 1

    logger.info(f"processed {n} FASTQ records from '{fname}'")


def make_header(bam, progname):
    import os
    import sys

    header = bam.header.to_dict()
    # if "PG" in header:
    # for pg in header['PG']:
    #     if pg["ID"] == progname:
    #         progname = progname + ".1"

    pg_list = header.get("PG", [])
    pg = {
        "ID": progname,
        "PN": progname,
        "CL": " ".join(sys.argv[1:]),
        "VN": __version__,
    }
    if len(pg_list):
        pg["PP"] = pg_list[-1]["ID"]

    header["PG"] = pg_list + [pg]
    return header


def dge_to_sparse(dge_path):
    import anndata
    import numpy as np
    import gzip
    import pandas as pd
    from scipy.sparse import csr_matrix, vstack

    ix = 0
    mix = 0
    matrices = []
    gene_names = []

    with gzip.open(dge_path, "rt") as dge:
        first_line = dge.readline().strip().split("\t")

        barcodes = first_line[1:]
        ncol = len(barcodes)
        M = np.zeros((1000, ncol))

        for line in dge:
            vals = line.strip().split("\t")
            # first element contains the gene name
            gene_names.append(vals[0])
            vals = vals[1:]
            vals = np.array(vals, dtype=np.int64)

            M[ix] = vals

            if ix % 1000 == 999:
                print(mix)
                mix = mix + 1
                matrices.append(csr_matrix(M))
                ix = 0
                M = np.zeros((1000, ncol))
            else:
                ix = ix + 1

        # get the leftovers
        M = M[:ix]
        matrices.append(csr_matrix(M))

        # sparse expression matrix
        X = vstack(matrices, format="csr")
        print(len(gene_names))

        adata = anndata.AnnData(
            X.T, obs=pd.DataFrame(index=barcodes), var=pd.DataFrame(index=gene_names)
        )

        return adata


def compute_neighbors(adata, min_dist=None, max_dist=None):
    """Compute all direct neighbors of all spots in the adata. Currently
        tailored for 10X Visium.
    Args:
        adata: an AnnData object
        min_dist: int, minimum distance to consider neighbors
        max_dist: int, maximum distance to consider neighbors
    Returns:
        neighbors: a dictionary holding the spot IDs for every spot
    """
    import numpy as np
    from scipy.spatial.distance import cdist

    # Calculate all Euclidean distances on the adata
    dist_mtrx = cdist(adata.obsm["spatial"], adata.obsm["spatial"])
    neighbors = dict()

    for i in range(adata.obs.shape[0]):
        neighbors[i] = np.where(
            (dist_mtrx[i, :] > min_dist) & (dist_mtrx[i, :] < max_dist)
        )[0]

    return neighbors


COMPLEMENT = {
    "a": "t",
    "t": "a",
    "c": "g",
    "g": "c",
    "k": "m",
    "m": "k",
    "r": "y",
    "y": "r",
    "s": "s",
    "w": "w",
    "b": "v",
    "v": "b",
    "h": "d",
    "d": "h",
    "n": "n",
    "A": "T",
    "T": "A",
    "C": "G",
    "G": "C",
    "K": "M",
    "M": "K",
    "R": "Y",
    "Y": "R",
    "S": "S",
    "W": "W",
    "B": "V",
    "V": "B",
    "H": "D",
    "D": "H",
    "N": "N",
    "-": "-",
    "=": "=",
    "+": "+",
}


def complement(s):
    return "".join([COMPLEMENT[x] for x in s])


def rev_comp(seq):
    return complement(seq)[::-1]


def fasta_chunks(lines, strip=True, fuse=True):
    chunk = ""
    data = []

    for l in lines:
        if l.startswith("#"):
            continue
        if l.startswith(">"):
            if data and chunk:
                # print chunk
                yield chunk, "".join(data)

                if strip:
                    data = []
                else:
                    data = [l]

            chunk = l[1:].strip()
        else:
            if fuse:
                data.append(l.rstrip())
            else:
                data.append(l)

    if data and chunk:
        yield chunk, "".join(data)


# @contextmanager
# def message_aggregation(log_listen="spacemake", print_logger=False, print_success=True):
#     message_buffer = []

#     log = logging.getLogger(log_listen)
#     log.setLevel(logging.INFO)

#     class MessageHandler(logging.NullHandler):
#         def handle(this, record):
#             if record.name == log_listen:
#                 if print_logger:
#                     print(f"{log_listen}: {record.msg}")
#                 else:
#                     print(record.msg)

#     log.addHandler(MessageHandler())

#     try:
#         yield True

#         if print_success:
#             print(f"{LINE_SEPARATOR}SUCCESS!")

#     except SpacemakeError as e:
#         print(e)


def message_aggregation(log_listen="spacemake", print_logger=False, print_success=True):
    from functools import wraps

    def the_decorator(func):
        @wraps(func)
        def wrapper(*argc, **kw):
            message_buffer = []

            log = logging.getLogger(log_listen)
            log.setLevel(logging.INFO)

            class MessageHandler(logging.NullHandler):
                def handle(this, record):
                    if record.name == log_listen:
                        if print_logger:
                            print(f"{log_listen}: {record.msg}")
                        else:
                            print(record.msg)

            log.addHandler(MessageHandler())

            try:
                res = func(*argc, **kw)
            except (SpacemakeError, NotImplementedError) as err:
                print(f"CAUGHT A SPACEMAKE ERROR {err}")
                # print(err)
                return err
            else:
                if print_success:
                    print(f"{LINE_SEPARATOR}SUCCESS!")
                return res

        return wrapper

    return the_decorator


def str_to_list(value):
    # if list in string representation, return the list
    if value is None:
        return None

    if type(value) is str and value.startswith("["):
        if value == "[nan]":
            return []
        else:
            return eval(value)
    # else create a list
    else:
        return [value]


def check_star_index_compatibility(star_index_dir):
    import os

    star_version = os.popen("STAR --version").read().strip()

    with open(os.path.join(star_index_dir, "Log.out"), "r") as f:
        first_line = f.readline().strip()
        index_version = first_line.split("=")[-1].split("_")[-1]

        if index_version != star_version:
            raise SpacemakeError(
                f"STAR index version ({index_version}) is"
                + f" incompatible with your STAR version ({star_version})"
            )


def load_yaml(path, mode="rt"):
    if not path:
        return {}
    else:
        import yaml

        return yaml.load(path, mode)


def setup_logging(
    args,
    name="spacemake.main",
    log_file="",
    FORMAT="%(asctime)-20s\t%(levelname)s\t{sample:20s}\t%(name)-30s\t%(message)s",
    rename_process=True,
    **kwargs,
):
    sample = getattr(args, "sample", kwargs.get("sample", "na"))
    if rename_process:
        import setproctitle

        if name != "spacemake.main":
            setproctitle.setproctitle(f"{name} {sample}")

    FORMAT = FORMAT.format(sample=sample)
    log_level = getattr(args, "log_level", "INFO")
    lvl = getattr(logging, log_level)
    # logging.basicConfig(level=lvl, format=FORMAT)
    # ^^ above does not mix well with Snakemake code
    root = logging.getLogger(name)
    root.setLevel(lvl)
    lh = logging.StreamHandler()
    lh.setFormatter(logging.Formatter(FORMAT))
    root.addHandler(lh)

    log_file = getattr(args, "log_file", log_file)
    if log_file:
        fh = logging.FileHandler(filename=ensure_path(log_file), mode="a")
        fh.setFormatter(logging.Formatter(FORMAT))
        root.debug(f"adding log-file handler '{log_file}'")
        root.addHandler(fh)

    if hasattr(args, "debug"):
        # cmdline requested debug output for specific domains (comma-separated)
        for logger_name in args.debug.split(","):
            if logger_name:
                root.info(f"setting domain {logger_name} to DEBUG")
                logging.getLogger(logger_name.replace("root", "")).setLevel(
                    logging.DEBUG
                )

    logger = logging.getLogger(name)
    logger.debug("started logging")
    for k, v in sorted(vars(args).items()):
        logger.debug(f"cmdline arg\t{k}={v}")

    return logger


def setup_smk_logging(name="spacemake.smk", **kw):
    import argparse

    args = argparse.Namespace(**kw)
    return setup_logging(args, name=name)


default_log_level = "INFO"


def make_minimal_parser(prog="", usage="", **kw):
    import argparse

    parser = argparse.ArgumentParser(prog=prog, usage=usage, **kw)
    parser.add_argument(
        "--log-file",
        default=f"",
        help=f"place log entries in this file (default=None)",
    )
    parser.add_argument(
        "--log-level",
        default=default_log_level,
        help=f"change threshold of python logging facility (default={default_log_level})",
    )
    parser.add_argument(
        "--debug",
        default="",
        help=f"comma-separated list of logging-domains for which you want DEBUG output",
    )
    parser.add_argument(
        "--sample", default="sample_NA", help="sample_id (where applicable)"
    )
    return parser


def load_config_with_fallbacks(args, try_yaml="config.yaml"):
    """
    Tries to load spacemake configuration from
        1) args.config
        2) try_yaml
        3) builtin default from spacamake package

    The entire configuration is attached to the args namespace such that
    args.config["barcode_flavors"]["dropseq] can work.
    """
    from spacemake.config import ConfigFile

    config = None
    if hasattr(args, "config") and os.access(args.config, os.R_OK):
        config = ConfigFile.from_yaml(args.config)
    elif os.access(try_yaml, os.R_OK):
        config = ConfigFile.from_yaml(try_yaml)
    else:
        builtin = os.path.join(os.path.dirname(__file__), "data/config/config.yaml")
        config = ConfigFile.from_yaml(builtin)

    args_kw = vars(args)
    # try:
    #     args_kw["log_level"] = config.variables["logging"]["level"]
    # except KeyError:
    #     pass

    args_kw["config"] = config.variables
    import argparse

    return argparse.Namespace(**args_kw)


def sync_timestamps(original_file, new_file):
    """
    Sync only the mtime of new_file to match original_file.
    For symlinks, mtime of the link itself is set (not the target).
    """
    import time

    try:
        source_stat = os.lstat(original_file)
        mtime = source_stat.st_mtime

        if os.path.islink(new_file):
            os.utime(new_file, (mtime, mtime), follow_symlinks=False)
        else:
            os.utime(new_file, (mtime, mtime), follow_symlinks=True)

    except Exception as e:
        print(f"[sync] Failed to sync timestamps for {new_file}: {e}")


def sync_symlink_mtime(source_symlink, target_symlink):
    """
    Set the mtime of a symlink to match another symlink using `touch -h`.
    Prints detailed debug info.
    """
    import subprocess
    import time

    try:
        # Format the timestamp in touch-compatible form: YYYYMMDDhhmm.ss
        mtime = os.lstat(source_symlink).st_mtime
        timestamp = time.strftime("%Y%m%d%H%M.%S", time.localtime(mtime))

        # Run touch
        result = subprocess.run(
            ["touch", "-h", "-t", timestamp, target_symlink],
            capture_output=True,
            text=True,
            check=False  # don't raise exception yet
        )

        if result.returncode != 0:
            print("[sync_symlink_mtime] touch failed.")
        else:
            print("[sync_symlink_mtime] touch succeeded.")

    except Exception as e:
        print(f"[sync_symlink_mtime] Exception occurred: {e}")




import pytest
import os

spacemake_dir = os.path.abspath(os.path.dirname(__file__) + "/../")
print("SPACEMAKE_DIR", spacemake_dir)


def sm(*argc, expect_fail=False):
    # construct the desired cmdline
    import sys

    sys.argv = [
        "spacemake",
    ] + list(argc)

    # ensure that no ConfigFile and ProjectDF instances
    # are retained from previous tests
    import spacemake.config
    import spacemake.project_df

    spacemake.config.__global_config = None
    spacemake.project_df.__global_ProjectDF = None

    # execute spacemake cmdline code
    from spacemake.cmdline import cmdline

    res = cmdline()
    # print("res", res)
    if expect_fail:
        assert isinstance(res, Exception) == True
    else:
        assert isinstance(res, Exception) == False

    return res


def _init():
    # just get the version
    sm("--version")

    # test the init parser
    sm("init", "--dropseq-tools", "/data/rajewsky/shared_bins/Drop-seq_tools-2.5.1/")


def _add_species():
    sm(
        "config",
        "add-species",
        "--name=test_hsa",
        "--reference=genome",
        f"--sequence={spacemake_dir}/test_data/simple_genome.fa",
        f"--annotation={spacemake_dir}/test_data/simple_annotation.gtf",
    )
    # add a second reference
    sm(
        "config",
        "add-species",
        "--name=test_hsa",
        "--reference=rRNA",
        f"--sequence={spacemake_dir}/test_data/rRNA_hsa.fa.gz",
    )
    # add a third reference
    sm(
        "config",
        "add-species",
        "--name=test_hsa",
        "--reference=miRNA",
        f"--sequence={spacemake_dir}/test_data/mirgenedb.hsa.mature.fa.gz",
    )
    # pretend we have mouse as well
    # TODO: place some actual mouse genome and/or phiX genomes in test-data repository
    sm(
        "config",
        "add-species",
        "--name=mouse",
        "--reference=genome",
        f"--sequence={spacemake_dir}/test_data/test_genome.fa.gz",
        f"--annotation={spacemake_dir}/test_data/test_genome.gtf.gz",
    )
    sm(
        "config",
        "add-species",
        "--name=mouse",
        "--reference=phiX",
        f"--sequence={spacemake_dir}/test_data/test_genome.fa.gz",
        f"--annotation={spacemake_dir}/test_data/test_genome.gtf.gz",
    )
    sm(
        "config",
        "add-species",
        "--name=mouse",
        "--reference=rRNA",
        f"--sequence={spacemake_dir}/test_data/rRNA_hsa.fa.gz",
    )


@pytest.fixture
def tmp_root(tmp_path_factory):
    tmp = tmp_path_factory.mktemp("root_blank")

    return tmp


@pytest.fixture
def initialized_root(tmp_path_factory):
    tmp = tmp_path_factory.mktemp("root_initialized")
    os.chdir(tmp.as_posix())

    _init()
    return tmp


@pytest.fixture
def with_species(initialized_root):
    os.chdir(initialized_root.as_posix())
    # # test old way
    # sm(
    #     "config", "add_species",
    #     "--name=hsa_test",
    #     f"--genome={spacemake_dir}/test_data/test_genome.fa.gz",
    #     f"--annotation={spacemake_dir}/test_data/test_genome.gtf.gz",
    # )
    # test new way
    _add_species()
    return initialized_root


@pytest.fixture
def configured_root(tmp_path_factory):
    tmp_root = tmp_path_factory.mktemp("root_preconfigured")

    # make a tmp-copy of the test_config.yaml
    def_config = os.path.join(spacemake_dir, "test_data/test_config.yaml")
    os.system(f"cp {def_config} {tmp_root / 'config.yaml'}")

    test_pdf = os.path.join(spacemake_dir, "test_data/test_project_df.csv")
    open(f"{tmp_root / 'project_df.csv'}", "w").write(
        open(test_pdf, "r").read().format(spacemake_dir=spacemake_dir)
    )
    # os.system(f"cp {test_pdf} {tmp_root / 'project_df.csv'}")

    return tmp_root


@pytest.fixture(scope="session")
def with_tile_test_data(tmp_path_factory):
    tmp = tmp_path_factory.mktemp("root_tile_test")
    os.chdir(tmp.as_posix())
    _init()
    _add_species()
    print(
        "return code",
        os.system(
            "wget https://bimsbstatic.mdc-berlin.de/rajewsky/spacemake-test-data/spacemake_tile_test_data.tar.gz -O /dev/stdout | tar -xz"
        ),
    )
    print(os.listdir("."))

    return tmp



# Configuration file for the Sphinx documentation builder.

# -- Project information

project = 'spacemake'
copyright = '2021-2024, Rajewsky Lab'
author = 'Tamas Ryszard Sztanka-Toth, Marvin Jens, Nikos Karaiskos, Nikolaus Rajewsky'

version = '0.9.1b'
release = version

# -- General configuration

extensions = [
    "sphinx_rtd_theme",
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosectionlabel',
    'nbsphinx'
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}

intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output
html_theme = "sphinx_rtd_theme"
html_theme_options = {
    'navigation_depth': 3
}

# -- Options for EPUB output
epub_show_urls = 'footnote'

import os
import sys
sys.path.insert(0, os.path.abspath('../'))



import os
import logging
from collections import OrderedDict, defaultdict
from spacemake.util import rev_comp

"""
Small helper class to load longread signature definitions (see docs/tutorials/longreads)
and make them accessible to the various cmdline tools.
"""

logger = logging.getLogger("spacemake.longread.signature")


class SignatureDB:
    def __init__(self, blocks=OrderedDict(), **kw):
        self.blocks = blocks
        self.lkup = {}
        self.fields = sorted(kw.keys())
        for f in self.fields:
            self.lkup[f] = kw[f]

    @classmethod
    def from_YAML(cls, fname="samples.yaml"):
        import yaml

        logger = logging.getLogger("spacemake.longread.SignatureDB.from_YAML")
        logger.info(f"reading longread signature definitions from '{fname}'")

        groups = yaml.load(open(fname), Loader=yaml.SafeLoader)
        signatures = groups["signatures"]
        default = signatures[groups["default"]]

        # load all building block oligo sequences and their reverse complements
        blocks = OrderedDict()
        for fa_id, seq in groups["blocks"].items():
            blocks[fa_id] = seq
            blocks[fa_id + "_RC"] = rev_comp(seq)

        logger.info(f"load_oligos(): loaded {len(blocks)} sequences from '{fname}'")

        # load the signature definitions and split into separate dictionaries
        field_lkups = {}
        for name, d in signatures.items():
            # print(f"name={name} d={d}")
            for f in d.keys():
                if f not in field_lkups:
                    field_lkups[f] = defaultdict(lambda: default[f])

                field_lkups[f][name] = d[f]

        logger.info(
            f"found {len(signatures)} signature definitions:"
            f"{sorted(signatures.keys())}."
        )
        return cls(blocks, **field_lkups)

    def __getattr__(self, attr):
        return self.lkup[attr]

    def sort_samples(self, samples, signatures):
        """
        Sort samples by the priority assigned in the signature definitions first,
        then lexicographically. Used for overview plots combining multiple longread
        sample results to group samples sharing a signature.
        """
        return sorted(
            zip(samples, signatures), key=lambda x: (self.prio.get(x[1], np.inf), x[0])
        )


def get_signature_db(try_path):
    """
    try to load a YAML file with longread signature definitions from <try_path>.
    If that fails, default to spacemake/data/config/longread.yaml
    """
    if os.access(try_path, os.R_OK):
        cfg = try_path
    else:
        cfg = os.path.join(os.path.dirname(__file__), "../data/config/longread.yaml")

    return SignatureDB.from_YAML(cfg)


def process_intact_signature(complete_signature, prefixes=["P5"], suffixes=["N70X"]):
    complete = complete_signature.split(",")
    while complete and complete[0] in prefixes:
        complete.pop(0)

    while complete and complete[-1] in suffixes:
        complete.pop()

    complete_order = dict(x[::-1] for x in enumerate(complete))
    # print(f"complete={complete}")

    return tuple(complete), complete_order


def digest_signatures(
    sig_counts,
    bead_related="bead_start",
    complete_signature="P5,bead_start,OP1,polyT,N70X",
    prefixes=[
        "P5",
    ],
    suffixes=[
        "N70X",
    ],
):
    bead_counts = defaultdict(int)
    ov_counts = defaultdict(int)
    n_bead_related = 0

    complete, complete_order = process_intact_signature(
        complete_signature, prefixes, suffixes
    )
    complete_set = set(complete)
    found_part_counts = defaultdict(int)

    def describe(found_set):
        missing = complete_set - found_set
        if not missing:
            descr = "complete"
        elif len(missing) < len(found_set):
            descr = f"missing_{','.join(sorted(missing))}"
        else:
            descr = f"only_{','.join(sorted(found_set))}"

        return descr

    def bead_relation(parts):
        search = list(complete)
        at = 0

        try:
            i = parts.index(search[0])  # look for first part, e.g. bead_start
        except ValueError:
            i = 0

        found = []
        for part in parts[i:]:
            # find co-linear matches,
            # ignore extra inserted segments
            # (for now)
            if part in search[at:]:
                found.append(part)
                at = search.index(part)

        found_set = set(found)
        found_tup = tuple(sorted(found_set, key=lambda x: complete_order[x]))

        return describe(found_set), found_tup

    for sig, count in sig_counts.items():
        parts = sig.split(",")
        if bead_related in parts:
            br, found_tup = bead_relation(parts)
            bead_counts[br] += count
            n_bead_related += count

            for i in range(1, len(found_tup) + 1):
                found_part_counts[found_tup[:i]] += count
        else:
            ov_counts[sig] = count

    ov_counts["bead-related"] = n_bead_related
    return ov_counts, bead_counts, found_part_counts, complete



# # include in top level for backward compatibility
# from .util import compute_neighbors, compute_islands, detect_tissue, \
#     create_mesh, create_meshed_adata
# # added novosparc_reconstruction for backward compatibility
# from . import novosparc_integration as novosparc_reconstruction
# from . import puck_collection as puck_collection



import pandas as pd
import argparse
import re

parser = argparse.ArgumentParser(
    description="Split .bam file to .sam files by mapped reads strand orientation"
)
parser.add_argument("file_in", metavar="in", type=str)
parser.add_argument("--prefix")

args = parser.parse_args()

prefix = args.prefix

read_type_num = {"INTERGENIC": 0, "INTRONIC": 0, "CODING": 0, "UTR": 0, "AMB": 0}

strand_type_num = {
    "minus_minus": 0,
    "minus_plus": 0,
    "plus_plus": 0,
    "plus_minus": 0,
    "plus_AMB": 0,
    "minus_AMB": 0,
}

# out_file_names = {x: prefix + x + ".sam" for x in strand_type_num.keys()}

# out_files = {x: open(out_file_names[x], "w") for x in out_file_names.keys()}


def return_collapsed(it):
    # set has exactly 1 element, meaning that all elements are the same in the list
    if len(set(it)) == 1:
        return it[0]
    else:
        return "AMB"


def get_tag_unique(line, tag="gs"):
    if m := re.search(f"{tag}\:Z\:(\S+)", line):
        gs = set(m.groups()[0].split(","))
        if len(gs) == 1:
            return gs.pop()


with open(args.file_in, "r") as fi:
    for line in fi:
        # if line is header line
        if line.startswith("@"):
            # for f in out_files.values():
            #     f.write(line)

            # # go to next iteration
            continue

        line_stripped = line.strip()

        elements = line_stripped.split()

        last = elements[-1]

        read_overlaps_gene = False

        # set gene strand
        if gene_strand := get_tag_unique(line_stripped, tag="gs"):
            # if last element begins with gs, this means that read overlaps a gene (fw or rv strands)
            read_overlaps_gene = True
            if gene_strand == "-":
                gene_strand = "minus"
            elif gene_strand == "+":
                gene_strand = "plus"

        else:
            gene_strand = "AMB"

        # print(f"gene_strand={gene_strand}")
        # set read strand
        if int(elements[1]) & 16:
            read_strand = "minus"
        else:
            read_strand = "plus"

        # get read type
        if read_overlaps_gene:
            if gf := get_tag_unique(line_stripped, tag="gf"):
                read_type = gf
            else:
                read_type = "AMB"
            # print(f"read_type={read_type}")
            # return_collapsed(elements[-3].split(":")[-1].split(","))
        else:
            # if read do not overlap a gene, it is clearly intergenic
            read_type = "INTERGENIC"

        read_type_num[read_type] = read_type_num[read_type] + 1

        strand_type = read_strand + "_" + gene_strand

        strand_type_num[strand_type] = strand_type_num[strand_type] + 1

        # # print the read to the correct split file, depending on strand orientation
        # out_files[strand_type].write(line)

with open(prefix + "read_type_num.txt", "w") as fo:
    for key, value in read_type_num.items():
        fo.write("%s %s\n" % (key, value))

with open(prefix + "strand_type_num.txt", "w") as fo:
    for key, value in strand_type_num.items():
        fo.write("%s %s\n" % (key, value))



import pandas as pd
import numpy as np
import logging
from spacemake.util import rev_comp
from spacemake.util import read_fq
from collections import defaultdict


def sig2str(sig, max_repeat_only=True):
    """
    basically merge oligo block names with ',' but contract homopolymeric repeats by
    adding a '+' suffix instead of repeating the thing over and over
    """
    if not len(sig):
        return ""

    parts = [sig[0]]
    poly_runs = set()
    repeat_counts = defaultdict(int)
    for s in sig[1:]:
        if s != parts[-1]:
            parts.append(s)
        else:
            poly_runs.add(len(parts) - 1)
            repeat_counts[s] += 1

    for i in poly_runs:
        parts[i] += "+"

    if poly_runs and max_repeat_only:
        # This is a more drastic reduction of the original signature: we drop
        # all other building blocks and disregard orientation. If any block X
        # is tandem repeated we call the signature of this just 'X+'.
        max_rep, nrep = sorted(repeat_counts.items(), key=lambda x: -x[1])[0]
        return f"?{max_rep.replace('_RC', '')}+"
    else:
        return ",".join(parts)


class AnnotatedSequences:
    def __init__(
        self,
        fastq_path,
        ann_path,
        sample_name,
        blocks,
        min_score=0,
        relevant=[],
        orient_by="bead_start",
    ):
        self.sample_name = sample_name
        self.logger = logging.getLogger("AnnotatedSequences")
        self.orient_by = orient_by
        self.orient_by_RC = orient_by + "_RC"
        self.raw_sequences = self.load_raw_sequences(fastq_path)
        self.oligo_blocks = blocks
        self.min_oligo_scores = {}
        self.relevant = set(relevant)
        for m in relevant:
            # reverse-complement hits of a relevant oligo too!
            self.relevant.add(f"{m}_RC".replace("_RC_RC", ""))

        for name, seq in blocks.items():
            self.min_oligo_scores[name] = (2 * len(seq)) * min_score

        self.logger.info(
            f"restricting to relevant matches {sorted(self.relevant)} "
            f"and discarding matches under {min_score} alignment score."
        )
        self.ann_db = self.load_annotation(ann_path)
        # self.ann_db = self.cleanup_overlaps(self.ann_db)

        self.signatures = self.extract_signatures_and_orient()

    def load_raw_sequences(self, fastq_path):
        self.logger.info(
            f"loading raw sequences for {self.sample_name} from {fastq_path}"
        )
        raw_sequences = {}
        for qname, seq, qual in read_fq(fastq_path):
            raw_sequences[qname] = seq

        return raw_sequences

    def load_annotation(self, ann_path):
        self.logger.info(
            f"loading oligo annotation for {self.sample_name} from {ann_path}"
        )
        df = pd.read_csv(ann_path, sep="\t")
        df["min_score"] = df["oligo"].apply(lambda x: self.min_oligo_scores.get(x, 22))
        # df = df.query(f"score > min_score")
        mask = df["score"] > df["min_score"]
        if len(self.relevant):
            mask &= df["oligo"].isin(self.relevant)

        df = df.loc[mask]

        qdata = {}
        for qname, grp in df.groupby("qname"):
            qdata[qname] = (
                tuple(grp["oligo"]),
                grp["start"].values,
                grp["end"].values,
                grp["score"].values,
            )

        return qdata

    def extract_signatures_and_orient(self):
        self.logger.info(f"extracting signatures for {self.sample_name}")
        signatures = {}
        for qname, (names, starts, ends, scores) in self.ann_db.items():
            seq = self.raw_sequences[qname]
            L = len(seq)
            lnames = list(names)
            fw_start = lnames.count(self.orient_by)
            rc_start = lnames.count(self.orient_by_RC)
            fw_nrc = lnames.count("_RC")
            rc_nrc = len(lnames) - fw_nrc

            if (
                (fw_start < rc_start)
                or ((fw_start == rc_start) and names[0].endswith("_RC"))
                or ((fw_start == rc_start) and (fw_nrc > rc_nrc))
            ):
                # reverse complement the whole thing
                names = tuple([(n + "_RC").replace("_RC_RC", "") for n in names[::-1]])
                scores = scores[::-1]
                starts, ends = L - ends[::-1], L - starts[::-1]
                seq = rev_comp(seq)
                # store the RC'ed data
                self.raw_sequences[qname] = seq
                self.ann_db[qname] = (names, starts, ends, scores)

            signatures[qname] = names

        return signatures

    def filter_signatures(self, qsig, substring=False):
        # print(f"qsig={qsig}")
        nmatch = 0
        qstr = ",".join(qsig)
        lq = len(qsig)
        for qname, sig in self.signatures.items():
            sstr = sig2str(sig)
            if substring:
                # demand match somewhere in signature (substring)
                i = sstr.find(qstr)
                if i > -1:
                    ofs = sstr[:i].count(",")
                    # print(qstr, sstr)
                    nmatch += 1
                    yield qname, ofs, ofs + lq
            else:
                # demand *exact* identity for a match
                # print(f"sig='{sig}' qsig='{qsig}'")
                if sstr == qstr:
                    nmatch += 1
                    yield qname, 0, lq

        self.logger.info(
            f"filter_signatures({qstr}, substring={substring}) -> {nmatch} ({100*nmatch/len(self.signatures):.2f}%) hits"
        )

    def count_signatures(self):
        sig_counts = defaultdict(int)
        concat = 0
        reprimed = 0
        for sig in self.signatures.values():
            sstr = sig2str(sig)
            sig_counts[sstr] += 1
            if "+" in sstr:
                concat += 1
            if len(sig) > 1:
                first = sig[0]
                last = sig[-1]
                if (first + "_RC").replace("_RC_RC", "") == last:
                    reprimed += 1

        return sig_counts, concat, reprimed

    def count_concatenations(self):
        concat = defaultdict(int)
        n_occ = 0
        for sig in self.signatures.values():
            sstr = sig2str(sig)
            nc = sstr.count("+")
            if nc:
                n_occ += nc
                for part in sstr.split(","):
                    if part.endswith("+"):
                        concat[part] += 1

        return concat, n_occ

    def count_repriming(self):
        reprimed = defaultdict(int)
        for sig in self.signatures.values():
            first = sig[0]
            last = sig[-1]
            first_rc = (first + "_RC").replace("_RC_RC", "")
            if first_rc == last:
                reprimed[first.replace("_RC", "")] += 1

        return reprimed

    def extract_cDNA(self, qname, after_oligo="bead_start", distal=150):
        """
        Ensure the read has a match to <after_oligo>. Find the last
        oligo-match following <after_oligo> and still within <distal> nt
        from start. The end of that match is the start of cDNA.
        On the other end, select the innermost oligo within <distal> nt from
        the end. The start of that match is the end of cDNA. Lastly, rev_comp
        the cDNA as Illumina would sequence from the 3'end.
        """
        read_seq = self.raw_sequences[qname]
        names, starts, ends, _ = self.ann_db[qname]
        if after_oligo in names:
            i = names.index(after_oligo)
        else:
            i = 0

        cDNA_start = 0
        L = len(read_seq)
        cDNA_end = L
        cDNA_end_oli = ""
        cDNA_start_oli = ""

        for name, start, end in zip(names[i:], starts[i:], ends[i:]):
            if start <= distal:
                cDNA_start = end
                cDNA_start_oli = name
            if end >= L - distal:
                cDNA_end = min(cDNA_end, start)
                cDNA_end_oli = name

        return (
            (cDNA_start, cDNA_end),
            (cDNA_start_oli, cDNA_end_oli),
            rev_comp(read_seq[cDNA_start:cDNA_end]),
        )

    def extract_between(
        self,
        qname,
        after="bead_start",
        before="polyT",
        srange=(0, np.inf),
        min_L=0,
        max_L=np.inf,
    ):
        """
        Useful for extracting barcodes and potentially also cDNA (TODO: unify).
        Looks for matches to <after> and <before> within the <range> interval of the
        oriented long read. Then excises the intervening sequence,
        if min_L <= length <= max_L .
        """

        read_seq = self.raw_sequences[qname]
        names, starts, ends, _ = self.ann_db[qname]

        extracted = ""
        s = None
        e = None
        if len(names) > 1:
            # iterate over consecutive pairs of detected oligo
            # matches and their start & end coordinates
            for (n1, s1, e1), (n2, s2, e2) in zip(
                zip(names[:-1], starts[:-1], ends[:-1]),
                zip(names[1:], starts[1:], ends[1:]),
            ):

                if s1 > srange[1] or e2 < srange[0]:
                    # outside the area we are scanning
                    continue

                if (n1 == after) and (n2 == before):
                    L = s2 - e1
                    if min_L <= L <= max_L:
                        s = e1
                        e = s2
                        extracted = read_seq[e1:s2]

                        break

        return extracted, (s, e)

    def query_dimensions(self, qsig, substring=False):
        qnames = []
        L = []
        starts = []
        ends = []
        scores = []

        for qname, i, j in self.filter_signatures(qsig, substring=substring):
            qnames.append(qname)
            L.append(len(self.raw_sequences[qname]))
            names, s, e, sc = self.ann_db[qname]
            starts.append(s[i:j])
            ends.append(e[i:j])
            scores.append(sc[i:j])

        return (
            np.array(qnames),
            np.array(L),
            np.array(starts),
            np.array(ends),
            np.array(scores),
        )

    def query_oligo_occurrences(self, oligo):
        qnames = []
        L = []
        starts = []
        ends = []
        scores = []

        for qname, sig in self.signatures.items():
            if oligo in sig:
                i = list(sig).index(oligo)

                qnames.append(qname)
                L.append(len(self.raw_sequences[qname]))
                names, s, e, sc = self.ann_db[qname]
                starts.append(s[i])
                ends.append(e[i])
                scores.append(sc[i])

        return (
            np.array(qnames),
            np.array(starts),
            np.array(ends),
            np.array(scores),
            np.array(L),
        )

    def fmt(self, qname):
        def render_label(label, start, end):
            return list(f"|{label.center(end - start - 2)}|")

        #
        seq = self.raw_sequences[qname]
        L = len(seq)
        buf = [" "] * L
        names, starts, ends, scores = self.ann_db[qname]
        # print(names)
        # print(starts)
        # print(ends)
        # print(scores)
        for oligo, start, end in zip(names, starts, ends):
            buf[start:end] = render_label(oligo, start, end)
        #
        return f"# qname={qname} oligo_matches={names} match_scores={scores}\n{seq}\n{''.join(buf)}"

    def completeness(self, sig_intact, polyT="polyT"):
        partial = []
        complete = ",".join(sig_intact)
        for i in range(len(sig_intact)):
            partial.append(",".join(sig_intact[: i + 1]))
            # print("partial append:", partial[-1])

        partial = partial[::-1]

        partial_counts = defaultdict(int)
        pT_counts = defaultdict(int)
        prefixes = defaultdict(int)
        suffixes = defaultdict(int)

        n = 0
        pT = f",{polyT}"
        # print(f"name of polyT oligo is '{polyT}' so scanning fot '{pT}'")
        for sig in self.signatures.values():
            n += 1
            sig_str = ",".join(sig)
            for i, p in enumerate(partial):
                # because we go in reverse order,
                # the first match is maximal. After that
                # we stop
                if p in sig_str:  #  and not (p + "_RC" in sig_str)
                    partial_counts[p] += 1
                    if pT in p:
                        # pT is part of the complete signature
                        pT_counts[p.rsplit(pT)[0]] += 1

                    assert partial_counts[p] <= n

                    if i == 0:
                        # if this is the maximal match (i==0)
                        # extract prefix and suffix
                        left_context = sig_str.split(p, maxsplit=1)
                        right_context = sig_str.rsplit(p, maxsplit=1)
                        prefixes[
                            left_context[0][:-1]
                        ] += 1  # [:-1] omits the trailing ','
                        if len(right_context) > 1:
                            if right_context[1].startswith(pT):
                                pT_counts[p] += 1
                            suffixes[
                                right_context[1][1:]
                            ] += 1  # [1:] omits the leading ','
                    break

        return partial_counts, prefixes, suffixes, pT_counts


def align_stats(ann, oseq, qmatches, pad=1, min_score=0):
    import spacemake.longread.cache as cache
    from Bio import pairwise2

    n = 0
    matches = np.zeros(len(oseq), dtype=float)
    edits = defaultdict(lambda: defaultdict(int))
    for qn, s, e, score, l in zip(*qmatches):
        s = max(0, s - pad)
        e = min(l, e + pad)
        n += 1
        rseq = ann.raw_sequences[qn][s:e]
        res = cache.align(rseq, oseq, min_score=min_score)
        if not res:
            continue
        aln = res[0]
        aln_str = pairwise2.format_alignment(*aln, full_sequences=True).split("\n")
        i = 0
        j = 0
        # if oseq.startswith("TTTTTTTTTTTTTTTT"):
        #     print("\n".join(aln_str))

        for r, m, q in zip(aln_str[0], aln_str[1], aln_str[2]):
            # print(r, m, q, i, len(oseq))
            if q == " ":
                # outside of oligo match range
                continue
            if m == " " and i == 0:
                # insertion at position zero is bananas for a local alignment
                # why do we even get this?
                continue
            if m == " " and i >= len(oseq):
                # same as above, we've already aligned all of the oligo sequence
                # what point is there in this "insertion"?
                continue
            if m == "|":
                # print(f"match {i}", matches[i])
                matches[i] += 1
            else:
                edits[i][q + r] += 1
                # print(f"edit {i} {q + r}")
            if q != "-":
                i += 1

    return matches / n, edits


def subsample(qmatches, n=100):
    data = np.array(qmatches, dtype=np.object)
    n_cols, N = data.shape
    I = np.random.choice(N, size=n, replace=False)
    return data[:, I]


def plot_mismatches(analyses, suffix=""):
    fig, ax = plt.subplots()
    fig2, axes = plt.subplots(len(analyses), figsize=(8, 12), sharex=True, sharey=True)
    if len(analyses) == 1:
        axes = [
            axes,
        ]
    #
    for (oname, region, matches), ax2 in zip(analyses, axes):
        if len(matches[0]) > 500:
            matches = subsample(matches, n=500)
        #
        qnames, starts, ends, scores, L = matches
        m, ed = align_stats(S, oname, matches)
        x = np.arange(len(m))
        ax.step(x, m, where="mid", lw=2, label=f"{oname} {region} n={len(qnames)}")
        #
        plot_edits_heatmap(ax2, oname, ed)
        ax2.title.set_text(f"{region} n={len(qnames)}")
        ax2.set_xlabel(f"{oname} sequence")
    #
    ax.set_xticks(x)
    ax.set_ylim(0.5, 1.0)
    ax.legend()
    ax.set_xticklabels(list(blocks[oname]))
    ax.set_ylabel("match frequency")
    ax.set_xlabel(f"{oname} oligo")
    fig.tight_layout()
    fig.savefig(f"match_{oname}_{sample}{suffix}.pdf")
    fig.savefig(f"match_{oname}_{sample}{suffix}.svg")
    fig2.suptitle(f"{sample} {oname} mismatch, deletion statistics by location")
    fig2.savefig(f"edits_{oname}_{sample}{suffix}.pdf")
    fig2.savefig(f"edits_{oname}_{sample}{suffix}.svg")
    plt.close("all")


#     stages[0] += 1
#     # print(sig)
#     l = list(sig)
#     if "SMART" in sig:
#         l = l[sig.index("SMART") + 1 :]
#         stages[1] += 1
#         try:
#             block = l.pop(0)
#             if block == "pT":
#                 pT[0] += 1
#             elif block == "OP1":
#                 stages[2] += 1
#                 block = l.pop(0)
#                 if block == "pT":
#                     pT[1] += 1
#                 elif block == "OP2":
#                     stages[3] += 1
#                     block = l.pop(0)
#                     if block == "pT":
#                         pT[2] += 1
#                     elif block == "OP3":
#                         stages[4] += 1
#                         block = l.pop(0)
#                         if block == "pT":
#                             pT[3] += 1
#         except IndexError:
#             # this just happens when l is empty
#             pass

# sig_intact = ("P5", "SMART", "OP1", "OP2", "OP3", "pT", "N70X")
# S = AnnotatedSequences(
#     sys.argv[1],
#     blocks,
#     path=base_path,
# )
# qintact, qL, qstarts, qends, qsores = S.query_dimensions(sig_intact)


# # estimate extension efficiencies
# # n_reads = 0
# stages = np.zeros(5, dtype=float)
# pT = np.zeros(4)
# for qname, sig in S.signatures.items():
#     stages[0] += 1
#     # print(sig)
#     l = list(sig)
#     if "SMART" in sig:
#         l = l[sig.index("SMART") + 1 :]
#         stages[1] += 1
#         try:
#             block = l.pop(0)
#             if block == "pT":
#                 pT[0] += 1
#             elif block == "OP1":
#                 stages[2] += 1
#                 block = l.pop(0)
#                 if block == "pT":
#                     pT[1] += 1
#                 elif block == "OP2":
#                     stages[3] += 1
#                     block = l.pop(0)
#                     if block == "pT":
#                         pT[2] += 1
#                     elif block == "OP3":
#                         stages[4] += 1
#                         block = l.pop(0)
#                         if block == "pT":
#                             pT[3] += 1
#         except IndexError:
#             # this just happens when l is empty
#             pass



import os
import sys
import logging
import pandas as pd
import numpy as np


def count_dict_collapse_misc(
    counts, misc_thresh=0.01, total=1, add_up=None, sig_intact=None
):
    out_counts = {}
    out_frac = {}

    misc = 0
    sum = 0
    if sig_intact is not None:
        complete = ",".join(sig_intact)
        everything = set(sig_intact)
    else:
        complete = None
        everything = set()

    def relkey(key):
        if sig_intact is None:
            return key

        if key == complete:
            return "complete"

        obs = set(key.split(","))
        there = obs & everything
        extra = obs - everything
        missing = everything - obs

        if len(missing) <= len(there):
            res = "missing_" + ",".join(sorted(missing))
        else:
            res = "only_" + ",".join(sorted(there))
        if extra:
            res += "_extra_" + ",".join(sorted(extra))

        return res

    for key, n in sorted(counts.items()):
        key = relkey(key)
        sum += n
        f = n / float(total)
        if f < misc_thresh:
            misc += n
        else:
            out_counts[key] = n
            out_frac[key] = f

        if misc > 0:
            out_counts["misc"] = misc
            out_frac["misc"] = misc / float(total)

    if add_up is None:
        other = total - sum
    else:
        other = total - counts[add_up]

    if other > 0:
        out_counts["NA"] = other
        out_frac["NA"] = other / float(total)
    return out_counts, out_frac


def count_dict_out(counts, title, misc_thresh=0.01, total=1, **kw):
    print(f"### {title}")
    out_counts, out_frac = count_dict_collapse_misc(counts, misc_thresh, total, **kw)
    for key in sorted(out_counts.keys()):
        print(f"{key}\t{out_counts[key]}\t{out_frac[key]:.3f}")


def to_hist(d, normed=True):
    x = np.array(list(d.keys()))
    x0 = x.min()
    x1 = x.max() + 1
    counts = np.zeros(x1, dtype=np.float32)

    for i in x:
        counts[i] = d[i]

    n = counts.sum()
    if normed:
        counts /= n

    return counts, n


def donut_plot(
    ax, data, sa=10, explode=None, colors=None, labels=None, title="", cmap="tab20"
):
    import matplotlib.pyplot as plt

    if labels is None:
        labels = sorted(data.keys())

    counts = [data.get(k, 0) for k in labels]

    if colors is None:
        colors = list(plt.cm.get_cmap(cmap)(np.linspace(0, 1, len(labels))))

    wedges, texts = ax.pie(
        counts,
        wedgeprops=dict(width=0.5),
        startangle=sa,
        explode=explode,
        colors=colors,
    )

    bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.5)
    kw = dict(arrowprops=dict(arrowstyle="-"), bbox=bbox_props, zorder=0, va="center")
    c = np.array(counts)
    pcts = 100.0 * c / float(c.sum())
    for i, p in enumerate(wedges):
        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        if pcts[i] > 0:
            ax.text(x * 0.75, y * 0.75, f"{pcts[i]:.1f}", horizontalalignment="center")
            ax.annotate(
                labels[i],
                xy=(x, y),
                xytext=(1.4 * np.sign(x), 1.4 * y),
                horizontalalignment=horizontalalignment,
                **kw,
            )

    if title:
        ax.set_title(title)

    return labels, colors


def approximate(intvalue):
    suffixes = {9: "G", 6: "M", 3: "k", 0: ""}
    dec = int(np.floor(np.log10(intvalue) / 3)) * 3
    x = np.round(intvalue / 10 ** dec, decimals=2)
    return f"{x:.2f} {suffixes.get(dec, '?')}"


def len_plot(
    ax,
    data,
    labels=None,
    colors=None,
    xlabel="aligned bases",
    ylabel="fraction",
    title="type",
    cmap="tab20",
    min_count=10,
    cumulative=False,
    legend=True,
):
    import matplotlib.pyplot as plt

    if labels is None:
        labels = sorted(data.keys())

    if colors is None:
        colors = plt.cm.get_cmap(cmap)(np.linspace(0, 1, len(labels)))

    color_dict = {}
    for cig_type, color in zip(labels, colors):
        color_dict[cig_type] = color

        if not cig_type in data:
            continue
        ml, n = to_hist(data[cig_type], normed=True)
        if n < min_count:
            continue

        x = np.arange(len(ml))
        y = ml.cumsum() if cumulative else ml

        ax.step(
            x,
            y,
            where="mid",
            label=f"{cig_type} ({approximate(n)})",
            color=color,
            lw=2,
            solid_capstyle="round",
        )

    if cumulative:
        ax.axhline(0.5, lw=0.5, ls="dashed", color="k")

    if legend:
        ax.legend(title=title, bbox_to_anchor=(0.5, 1.05), loc="lower center", ncol=2)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return color_dict


# def make_colors_explode(labels, cmap="Blues", hilight="bead-related", hicolor="red"):
#     import matplotlib.pyplot as plt
#     ex = np.zeros(len(labels))
#     colors = list(plt.get_cmap(cmap)(np.linspace(0.2, 0.8, len(labels))))
#     try:
#         i = labels.index(hilight)
#     except ValueError:
#         pass
#     else:
#         ex[i] = 0.1
#         colors[i] = hicolor
#     return ex, colors



#!/usr/bin/env python3
__version__ = "0.9"
__author__ = [
    "Marvin Jens",
]
__license__ = "GPL"
__email__ = [
    "marvin.jens@mdc-berlin.de",
]

import pysam
import argparse
import os
import sys
import logging


def print_header(header):
    for k, v in sorted(header.items()):
        if type(v) == dict:
            vstr = " ".join([f"{x}:{y}" for x, y in sorted(v.items())])
            print(f"@{k}:\t{vstr}")
        elif type(v) == list:
            for row in v:
                if type(row) == dict:
                    vstr = " ".join([f"{x}:{y}" for x, y in sorted(row.items())])
                else:
                    vstr = str(row)
                print(f"@{k}:\t{vstr}")
        else:
            print(f"@{k}:\t{v}")


def merge_headers(orig, star):
    merged = dict(orig)
    # most recent program should be on top
    merged["PG"] = star["PG"] + merged["PG"]
    merged["SQ"] = star["SQ"]
    merged["HD"]["SO"] = star["HD"]["SO"]  # sorted by

    return merged


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Fix .bam header of the STAR mapped output .bam"
    )

    parser.add_argument("--in-bam-star", help="mapped star bam input")
    parser.add_argument("--in-bam-tagged", help="unmapped dropseq tagged bam")
    parser.add_argument("--out-bam", help="output bam")

    args = parser.parse_args()

    bam_star = pysam.AlignmentFile(args.in_bam_star, "rb")
    bam_tagged = pysam.AlignmentFile(args.in_bam_tagged, "rb", check_sq=False)

    star_header = bam_star.header.to_dict()
    tagged_header = bam_tagged.header.to_dict()
    merged_header = merge_headers(tagged_header, star_header)
    # print(f"STAR header")
    # print_header(star_header)

    # print(f"original header")
    # print_header(tagged_header)

    # print("merged header")
    # print_header(merged_header)

    # copy input to output, just with the new header
    bam_out = pysam.AlignmentFile(args.out_bam, "wb", header=merged_header)
    for aln in bam_star.fetch(until_eof=True):
        bam_out.write(aln)



import numpy as np
import pandas as pd
import scanpy as sc
from spacemake.util import detect_tissue, attach_barcode_file

dge_path = snakemake.input['dge']

# umi cutoff 
umi_cutoff = int(snakemake.wildcards['umi_cutoff'])

adata = sc.read_h5ad(dge_path)

print('data read')

has_barcode_file = 'barcode_file' in snakemake.input.keys()

# ATTACH BARCODE FILE #
if has_barcode_file:
    adata = attach_barcode_file(adata, snakemake.input['barcode_file'])

# filter out cells based on umi, and genes based on number of cells
sc.pp.filter_cells(adata, min_genes=1)
sc.pp.filter_genes(adata, min_cells=3)

print('data filtered')

# DETECT TISSUE #
# if there is no barcode file, filter adata based on UMI, otherwise detect tissue with UMI cutoff
if has_barcode_file and snakemake.params['downstream_variables']['detect_tissue']:
    tissue_indices = detect_tissue(adata, umi_cutoff)
    adata = adata[tissue_indices, :]
else:
    adata = adata[adata.obs.total_counts > umi_cutoff, :]

adata.write(snakemake.output[0])



import os
import numpy as np
import logging
from collections import defaultdict

import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt


def multi_row_barplots(dfs, row_labels, samples, attr, color_dict={}):
    import spacemake.longread.util as util

    fig, axes = plt.subplots(len(row_labels), sharex=True, sharey=True, figsize=(10, 8))
    for df, part, ax in zip(dfs, row_labels, axes):
        x, y = util.gather_data_from_overview(df, samples, attr)
        c = [color_dict[s] for s in samples]
        ax.bar(x, y, color=c)
        ax.set_ylabel(part)
        ax.set_xticks(x)
        ax.set_xticklabels(samples, rotation=45, horizontalalignment="right")

        for _x, _y in zip(x, y):
            ax.text(
                _x,
                0.5,
                f"{_y:.2f}",
                horizontalalignment="center",
                verticalalignment="center",
            )
    return fig, axes


def donut_plot(
    ax,
    labels,
    counts,
    sa=10,
    explode=None,
    colors=None,
    w=0.5,
    radius=1,
    label_inside=False,
    **kw,
):
    wedges, texts = ax.pie(
        counts,
        wedgeprops=dict(width=w * radius),
        startangle=sa,
        explode=explode,
        colors=colors,
        radius=radius,
        **kw,
    )
    bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.5)
    kw = dict(arrowprops=dict(arrowstyle="-"), bbox=bbox_props, zorder=0, va="center")
    c = np.array(counts)
    pcts = 100.0 * c / float(c.sum())
    for i, p in enumerate(wedges):
        ang = (p.theta2 - p.theta1) / 2.0 + p.theta1
        y = np.sin(np.deg2rad(ang)) * radius
        x = np.cos(np.deg2rad(ang)) * radius
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        # print(i, labels[i], pcts[i], x, y)
        if pcts[i] > 1.0:
            if label_inside:
                label = f"{labels[i]}\n{pcts[i]:.1f}"
            else:
                label = f"{pcts[i]:.1f}"

            # draw text inside the wedge
            ax.text(
                x * (radius - w / 2),  # * radius,
                y * (radius - w / 2),  # * radius,
                label,
                horizontalalignment="center",
                verticalalignment="center",
            )
            # draw annotation label outside, linked with solid line
            if not label_inside:
                ax.annotate(
                    labels[i],
                    xy=(x, y),
                    xytext=((radius + 0.15) * np.sign(x), y * 1.1),
                    horizontalalignment=horizontalalignment,
                    **kw,
                )


def make_colors_explode(
    labels, cmap="Blues", hilight="bead-related", hicolor="red", color_order=[]
):
    ex = np.zeros(len(labels))
    colors = list(plt.get_cmap(cmap)(np.linspace(0.2, 0.8, len(labels))))
    # assigned = ['gray'] * len(labels)
    # for i, l in enumerate(labels):
    #     if l in color_order:
    #         assigned[i] = colors[color_order.index(l)]
    # for j, name in enumerate(color_order):
    #     try:
    #         z = labels.index(name)
    #     except ValueError:
    #         pass
    #     else:
    #         colors[z], colors[j] = colors[j], colors[z]
    try:
        i = labels.index(hilight)
    except ValueError:
        pass
    else:
        ex[i] = 0.1
        colors[i] = hicolor
    return ex, colors


def overview_plots(df, path=".", bead=[], concatenation=[], repriming=[]):
    def make_bars(
        ax, df, kinds, labels, cmap=plt.get_cmap("tab10"), w=0.9, colors=None
    ):
        n = len(kinds)
        if colors is None:
            colors = cmap(np.linspace(0, 1, n))

        x = np.arange(len(df)) - w / 2.0
        y0 = np.zeros(len(x), dtype=float)
        for kind, label, color in zip(kinds, labels, colors):
            y = np.nan_to_num(df[kind], nan=0.0)
            # print(kind)
            # print(y)
            ax.bar(x, y, bottom=y0, label=label, width=w, color=color)
            y0 += y

        ax.set_ylabel("fraction of library")
        ax.set_xticks(x)
        labels = df["sample"]  # [clean(fq) for fq in df['qfa']]
        ax.set_xticklabels(labels, rotation=90)
        ax.set_ylim(0, 100)

    marie = [
        "non_bead",
        "incomplete",
        "dropseq",
        "complete",
    ]
    marie_colors = ["gray", "royalblue", "green", "gold"]

    ## First plot: fidelity.pdf
    w = max(8 / 25.0 * len(df), 5)
    fig, (ax1, ax2) = plt.subplots(2, figsize=(w, 8), sharex=True)

    make_bars(
        ax1,
        df,
        marie,
        labels=marie,
        colors=marie_colors,
    )
    ax1.legend(title="Marie-stats", ncol=len(marie))
    make_bars(ax2, df, ["fidelity"], labels=["bead fidelity"])

    ax2.set_ylabel("bead fidelity")
    fig.tight_layout()
    fig.savefig(os.path.join(path, "fidelity.pdf"))

    ## Second plot: artifacts.pdf
    fig, (ax1, ax2, ax3) = plt.subplots(3, figsize=(w, 10), sharex=True)
    make_bars(ax1, df, bead, labels=bead)
    ax1.legend(title="bead-related", ncol=len(bead))

    # print("repriming events", repriming)
    make_bars(
        ax2,
        df,
        repriming,
        labels=[r.split(",")[0] for r in repriming],
        cmap=plt.get_cmap("tab20c"),
    )
    ax2.legend(title="repriming", ncol=len(repriming))

    # print("concat events", concatenation)
    make_bars(ax3, df, concatenation, labels=concatenation, cmap=plt.get_cmap("tab20b"))
    ax3.legend(title="concatamers", ncol=len(concatenation))

    fig.tight_layout()
    fig.savefig(os.path.join(path, "artifacts.pdf"))


def plot_results(
    sig_counts,
    donut_labels,
    donut_counts,
    bead_normed_labels,
    bead_normed_counts,
    syn_rates,
    all_parts=["bead_start", "polyT"],
    fname="donuts.pdf",
    suptitle="",
):

    fig, axes = plt.subplots(2, 2, figsize=(12, 8))
    [(ax1, ax2), (ax0, ax3)] = axes
    n_total = sig_counts["n_total"]

    if suptitle:
        fig.suptitle(f"{suptitle} (n_reads={n_total})")

    ax0.set_title("top 10 read signatures")
    ax0.set_xscale("log")

    # Hide the right and top spines
    ax0.spines["right"].set_visible(False)
    ax0.spines["top"].set_visible(False)

    # Only show ticks on the left and bottom spines
    sig_items = sorted(sig_counts.items(), key=lambda x: x[1])[-10:]
    sig_labels = [f"{x[0]} ({x[1]})" for x in sig_items]
    sig_cnt = [x[1] for x in sig_items]
    y_pos = np.arange(len(sig_labels))
    ax0.barh(y_pos, sig_cnt, height=0.2)
    ax0.set_yticks(y_pos)
    ax0.set_yticklabels(sig_labels)

    ax1.set_title("library overview")
    ax1.set_aspect("equal")
    ex, colors = make_colors_explode(
        donut_labels, cmap="GnBu", hilight="bead-related", hicolor="lawngreen"
    )
    donut_plot(ax1, donut_labels, donut_counts, sa=40, explode=ex, colors=colors)

    ax2.set_title("bead completeness")
    ax2.set_aspect("equal")
    ex, colors = make_colors_explode(
        bead_normed_labels,
        cmap="YlOrRd",
        hilight="complete",
        hicolor="seagreen",
        color_order=["missing_OP1", "missing_pT", "only_bead_start"],
    )
    donut_plot(
        ax2, bead_normed_labels, bead_normed_counts, sa=30, explode=ex, colors=colors
    )
    # estimate synthesis completion rates based on co-linear detection
    # of the constitutive parts of the capture oligos
    ax3.set_title("synthesis rates")
    ax3.spines["right"].set_visible(False)
    ax3.spines["top"].set_visible(False)
    x = range(1, len(syn_rates) + 1)
    ax3.barh(x, syn_rates[::-1])
    ax3.set_yticks(x)
    ax3.set_yticklabels([all_parts[i - 1] for i in x[::-1]])
    ax3.set_xlabel("synthesis completion")
    ax3.set_xticks([0, 0.2, 0.4, 0.6, 0.8, 1.0])
    ax3.set_xlim(0, 1)

    fig.subplots_adjust(
        top=0.942, bottom=0.16, left=0.125, right=0.813, hspace=0.167, wspace=0.2
    )

    # fname = os.path.basename(sys.argv[1]).replace('.fq', '.report')
    logger = logging.getLogger("pb_annotate.plot_results")
    logger.info(f"saving donut-charts with annotation breakdown to '{fname}'")
    fig.tight_layout(rect=[0, 0.03, 1, 0.95])
    plt.savefig(fname)

    return fig, axes


def obs_to_arrays(df):
    x = df["value"].values
    y = df["freq"].values
    return x, y


def plot_histograms(df, fname, parts=["bead_start", "OP1", "pT"], n_total=1):
    import seaborn as sns

    all_ends = df.query("attr == 'end'")["value"]
    max_x = int((np.percentile(all_ends, 95) + 1) / 100) * 100

    print(
        f"position 95th percentile would be {np.percentile(all_ends, 95)} -> max_x={max_x}"
    )

    n_parts = len(parts)
    fig, axes = plt.subplots(4, n_parts, figsize=(2 + n_parts * 3, 8))
    # axes = np.array(axes)
    # print(axes, n_parts)
    # first row: start coordinates, second row end coordinates, third row score distributions
    colors = plt.get_cmap("tab10")(np.arange(len(parts)))
    for ax_row, attr in zip(axes, ["start", "end", "score", "len"]):
        for part, ax, color in zip(parts, ax_row, colors):
            # ax.set_yscale("log")
            ctrl_x, ctrl_y = obs_to_arrays(
                df.query(
                    f"signature == 'anywhere' and oligo == '{part}' and attr == '{attr}'"
                )
            )
            df_sig = df.query(
                f"signature != 'anywhere' and oligo == '{part}' and attr == '{attr}'"
            )
            # print(df_sig)
            sig_x, sig_y = obs_to_arrays(df_sig)

            if attr == "len":
                ys = sig_y / n_total
                yc = ctrl_y / n_total
            else:
                ys = sig_y.cumsum() / n_total
                yc = ctrl_y.cumsum() / n_total

            ax.plot(
                sig_x,
                ys,
                "--",
                color=color,
                label=f"intact",
            )
            ax.plot(
                ctrl_x,
                yc,
                "-",
                color=color,
                label=f"all",
            )
            if attr == "start":
                ax.set_title(part)
            ax.legend(frameon=False)
            ax.set_xlabel(f"match {attr}")
            if attr != "len":
                ax.set_ylabel("cumulative rel. frequency")
                ax.set_ylim(0, 1)
            else:
                ax.set_ylabel("rel. frequency")
            if attr in ["start", "end"]:
                ax.set_xlim(0, max_x)

    fig.tight_layout()
    fig.savefig(fname)


def plot_edits_heatmap(ax, oname, oseq, edits, nmatch=1):
    N = len(oseq)
    counts = np.zeros((N, 6), dtype=float)
    x = np.arange(N)
    for i in x:
        for j, nn in enumerate("ACGT-"):
            if nn == oseq[i]:
                counts[i, j] = np.nan
            else:
                counts[i, j] = edits.iloc[i][oseq[i] + nn]

        is_ins = np.array([edits.iloc[i]["-" + ins] for ins in "ACGT"]).sum()
        counts[i, 5] = is_ins

    # print(counts, "non-finite entries", (~np.isfinite(counts)).sum())
    # print(counts / nmatch)
    im = ax.imshow(counts.T / nmatch, interpolation="none", cmap="viridis")
    plt.colorbar(im, ax=ax, shrink=0.5)
    ax.set_xticks(x)
    ax.set_xticklabels(list(oseq))
    ax.set_yticks(range(6))
    ax.set_yticklabels(list("ACGT-+"))


def plot_edits(df, fname, parts=["bead_start", "OP1", "pT"]):
    n_parts = len(parts)
    fig, axes = plt.subplots(n_parts, 2, figsize=(10, 1 + n_parts * 2))
    colors = plt.get_cmap("tab10")(np.arange(len(parts)))
    for part, (ax1, ax2), color in zip(parts, axes, colors):
        dfp = df.query(f"oligo == '{part}'")
        # print(dfp["pos"], dfp["fmatch"])
        ax1.step(dfp["pos"].values, dfp["fmatch"].values, where="mid", lw=2)
        ax1.set_xlabel(f"{part} pos. [nt]")
        ax1.set_ylabel("match frequency")
        #
        if len(dfp["seq"]):
            plot_edits_heatmap(
                ax2,
                part,
                dfp["seq"].iloc[0],
                dfp["ed_dict"],
                nmatch=df["nmatch"].iloc[0],
            )
            # ax2.title.set_text(f"{region} n={len(qnames)}")
            ax2.set_xlabel(f"{part} sequence")
        ax1.set_ylim(0, 1)

    fig.tight_layout()
    fig.savefig(fname)


def count_dict_collapse_misc(
    counts, misc_thresh=0.01, total=1, add_up=None, sig_intact=None
):
    out_counts = defaultdict(int)
    out_frac = defaultdict(float)

    misc = 0
    sum = 0
    if sig_intact is not None:
        complete = ",".join(sig_intact)
        everything = set(sig_intact)
    else:
        complete = None
        everything = set()

    def relkey(key):
        if sig_intact is None:
            return key

        if key == complete:
            return "complete"

        obs = set(key.split(","))
        there = obs & everything
        extra = obs - everything
        missing = everything - obs

        if len(missing) <= len(there):
            res = "missing_" + ",".join(sorted(missing))
        else:
            res = "only_" + ",".join(sorted(there))
        if extra:
            res += "_extra_" + ",".join(sorted(extra))

        return res

    for key, n in sorted(counts.items()):
        key = relkey(key)
        sum += n
        f = n / float(total)
        if f < misc_thresh:
            misc += n
        else:
            out_counts[key] = n
            out_frac[key] = f

    if misc > 0:
        out_counts["misc"] += misc
        out_frac["misc"] += misc / float(total)

    if add_up is None:
        other = total - sum
    else:
        other = total - counts[add_up]

    if other > 0:
        out_counts["N/A"] += other
        out_frac["N/A"] += other / float(total)

    return out_counts, out_frac


def count_dict_out(counts, title, misc_thresh=0.01, total=1, **kw):
    print(f"### {title}")
    colname = title.replace(" ", "_")
    out_counts, out_frac = count_dict_collapse_misc(counts, misc_thresh, total, **kw)
    for key, count in sorted(out_counts.items(), key=lambda x: -x[1]):
        print(f"{colname}\t{key}\t{count}\t{out_frac[key]:.3f}")

    return out_counts, out_frac


def count_dict_split(counts, pattern, name):
    import re

    out_d = defaultdict(int)
    for key, value in counts.items():
        if re.search(pattern, key):
            out_d[name] += value
        else:
            out_d[key] += value

    return out_d


def count_dict_to_df(counts, kind="", n_total=0):
    keys_values = sorted(counts.items(), key=lambda x: -x[1])
    import pandas as pd

    df = pd.DataFrame(data=keys_values, columns=["name", "count"])
    if n_total:
        df = df.append({"name": "n_total", "count": n_total}, ignore_index=True)
        df["fraction"] = df["count"] / n_total
    if kind:
        df["kind"] = kind
    return df


def count_dict_from_df(df, kind):
    df = df.query(f"kind == '{kind}'")
    # print(df)
    keys = df["name"]
    values = df["count"]
    return dict(zip(keys, values))


def gather_data_from_overview(df, samples, attr, na=np.nan):
    # print(df)
    # maybe not the most elegant, but doesn't crash on missing values.
    # TODO: look at pivot for a cleaner implementation
    x = np.arange(len(samples))
    # print(attr, samples)
    raw = [df.query(f"sample == '{s}'")[attr].values for s in samples]
    y = [r[0] if len(r) else na for r in raw]

    return x, y



__version__ = "0.9.6"
__author__ = ["Marvin Jens"]
__license__ = "GPL"
__email__ = ["marvin.jens@mdc-berlin.de"]

import os
import sys
import logging
import pandas as pd
import numpy as np
import pysam
import argparse
from collections import defaultdict

import spacemake.reporting as rep

# TODO:
# * graphical output of most common soft-clipped sequences
#   (MUSCLE?)


def gf_prio(gf, hierarchy=["CODING", "UTR", "INTRONIC", "INTERGENIC"]):
    parts = gf.split(",")
    if len(parts) == 1:
        # fast path if only one gf is given
        return gf

    parts = set(parts)
    for h in hierarchy:
        if h in parts:
            return h

    # could not resolve??
    return gf


def parse_args():
    parser = argparse.ArgumentParser("alnstats")
    parser.add_argument("fname", help="a gene-annotated BAM file")
    parser.add_argument(
        "--skim",
        type=int,
        default=0,
        help="only examin every n-th alignment (default=0 -> all)",
    )
    parser.add_argument(
        "--intact-signature",
        type=str,
        default="",
        help="use for longread cDNA: analyze separately alignments with and without intact-signature in the read-name",
    )
    parser.add_argument(
        "--parse-oligos",
        default=False,
        action="store_true",
        help="use for longread cDNA: calculate association of oligo-presence/absence with map status and gene annotation",
    )

    parser.add_argument(
        "--out-csv", default="./", help="where to store the CSV raw stats"
    )
    parser.add_argument(
        "--out-pdf", default="./", help="where to store the PDF reports"
    )
    parser.add_argument(
        "--out-png", default="./", help="where to store the PNG reports"
    )
    args = parser.parse_args()

    return args


class Results:
    pass


def coarsegrain_CIGAR(cigar):
    """
    recursively remove insertions, deletions, and gaps (introns) (I,D,N)
    from CIGAR string. Keep only matching and soft-clipping info (M,S)
    """
    coarse = cigar.replace("MNM", "M").replace("MDM", "M").replace("MIM", "M")
    if coarse == cigar:
        return cigar
    else:
        return coarsegrain_CIGAR(coarse)


def scan_bam(fname, skim=0, intact_signature="", parse_oligos=False):
    res = Results()
    res.fname = fname
    res.skim = skim
    res.aln_types = defaultdict(int)
    res.cigar_types = defaultdict(int)
    res.tag_types = defaultdict(int)
    res.uniq_tag_types = defaultdict(int)
    res.match_len_by_cigtype = defaultdict(lambda: defaultdict(int))
    res.match_len_by_tag = defaultdict(lambda: defaultdict(int))
    res.match_len_by_utag = defaultdict(lambda: defaultdict(int))
    res.sc_lengths = defaultdict(lambda: defaultdict(int))
    res.sc_seq = defaultdict(lambda: defaultdict(int))
    res.read_lengths_by_mapstat = defaultdict(lambda: defaultdict(int))
    res.oligo_by_mapstat = defaultdict(lambda: defaultdict(int))
    res.oligo_by_tag = defaultdict(lambda: defaultdict(int))
    res.oligo_count = defaultdict(int)
    res.all_oligos = set()

    op_dict = list("MIDNSHP=XB")
    sam = pysam.Samfile(fname, "rb")
    for i, r in enumerate(sam.fetch(until_eof=True)):
        if skim:
            if (i % skim) != 0:
                continue

        if intact_signature:
            if not intact_signature in r.qname:
                continue

        parts = []
        if parse_oligos:
            try:
                parts = set(r.qname.split("__sig:")[1].split("__")[0].split(","))
            except (ValueError, KeyError):
                pass

        res.aln_types["N_reads"] += 1
        for p in parts:
            res.oligo_count[p] += 1
            res.all_oligos.add(p)

        l_read = len(r.query_sequence)
        res.read_lengths_by_mapstat["all"][l_read] += 1

        if r.is_unmapped:
            res.aln_types["unmapped"] += 1
            res.read_lengths_by_mapstat["unmapped"][l_read] += 1
            for p in parts:
                res.oligo_by_mapstat["unmapped"][p] += 1
                res.oligo_by_tag["NA"][p] += 1
            continue

        minus_strand = r.is_reverse
        if r.has_tag("gf"):
            tag = gf_prio(r.get_tag("gf"))
        else:
            tag = "INTERGENIC"

        res.tag_types[tag] += 1

        is_uniq = r.mapping_quality == 255
        if not is_uniq:
            mapstat = "multimapper"
        else:
            mapstat = "unique"
            res.uniq_tag_types[tag] += 1

        for p in parts:
            res.oligo_by_mapstat[mapstat][p] += 1
            res.oligo_by_tag[tag][p] += 1

        res.aln_types[mapstat] += 1
        res.read_lengths_by_mapstat[mapstat][l_read] += 1
        if mapstat in ["multimapper", "unique"]:
            res.aln_types["mapped"] += 1
            res.read_lengths_by_mapstat["mapped"][l_read] += 1
            for p in parts:
                res.oligo_by_mapstat["mapped"][p] += 1

        cigar = list(r.cigartuples)
        cigtype = "".join([op_dict[c[0]] for c in cigar])
        cigtype = coarsegrain_CIGAR(cigtype)  # ignore splicing
        res.cigar_types[cigtype] += 1

        if cigar[0][0] == pysam.CSOFT_CLIP:
            n = cigar[0][1]
            seq = r.query[:n]
            if minus_strand:
                res.sc_seq["3'"][seq] += 1
                res.sc_lengths["3'"][n] += 1
            else:
                res.sc_seq["5'"][seq] += 1
                res.sc_lengths["5'"][n] += 1

        if cigar[-1][0] == pysam.CSOFT_CLIP:
            n = cigar[-1][1]
            seq = r.query[-n:]
            if minus_strand:
                res.sc_seq["5'"][seq] += 1
                res.sc_lengths["5'"][n] += 1
            else:
                res.sc_seq["3'"][seq] += 1
                res.sc_lengths["3'"][n] += 1

        n_match = 0
        for op, n in r.cigartuples:
            if op == pysam.CMATCH:
                n_match += n

        res.match_len_by_cigtype[cigtype][n_match] += 1
        res.match_len_by_cigtype["all"][n_match] += 1

        res.match_len_by_tag[tag][n_match] += 1
        res.match_len_by_tag["all"][n_match] += 1

        if is_uniq:
            res.match_len_by_utag[tag][n_match] += 1
            res.match_len_by_utag["all"][n_match] += 1

    res.N_total = res.aln_types["N_reads"]
    return res


def make_plots(res, required_aln_types=["unmapped", "mapped", "multimapper", "unique"]):
    import matplotlib.pyplot as plt

    # preparing count dictionaries for the donut plots (pie charts)
    aln_types_d, _ = rep.count_dict_collapse_misc(
        res.aln_types, total=res.N_total, add_up="N_reads"
    )
    # ensure that these keys are always present!
    for t in required_aln_types:
        aln_types_d[t] = res.aln_types.get(t, 0)

    # cig_types_d, _ = rep.count_dict_collapse_misc(res.cigar_types, total=res.N_total)
    tag_types_d, _ = rep.count_dict_collapse_misc(res.tag_types, total=res.N_total)
    utag_types_d, _ = rep.count_dict_collapse_misc(
        res.uniq_tag_types, total=res.aln_types["unique"]
    )

    # preparing the plot
    fig, axes = plt.subplots(
        3, 3, figsize=(14, 12)
    )  # gridspec_kw={'height_ratios': [1, 1, 1]})
    axes = np.array(axes).flatten()

    # # First row: donut plots
    # ciglabels, cigcolors = rep.donut_plot(
    #     axes[0], cig_types_d, title="CIGAR types", cmap="Accent"
    # )
    alnlabels, alncolors = rep.donut_plot(
        axes[0],
        aln_types_d,
        title="Alignment status",
        labels=["unique", "multimapper", "unmapped"],
        cmap="Set2",
    )

    taglabels, tagcolors = rep.donut_plot(
        axes[1], tag_types_d, title="annotation", cmap="Dark2"
    )

    utaglabels, utagcolors = rep.donut_plot(
        axes[2], utag_types_d, title="annotation (uniq only)", cmap="Dark2"
    )
    mean_lens_d = {}
    for mapstat, lens in res.read_lengths_by_mapstat.items():
        # print(mapstat, lens)
        l, f = np.array(list(lens.items())).T
        mean_lens_d[mapstat] = (l * f).sum() / f.sum()

    print(mean_lens_d)

    # Second row: histograms
    rep.len_plot(
        axes[3],
        res.read_lengths_by_mapstat,
        title="alignment status",
        labels=alnlabels + ["all"],
        colors=alncolors + ["k"],
        xlabel="read length [nt]",
    )

    rep.len_plot(
        axes[4],
        res.match_len_by_tag,
        labels=taglabels + ["all"],
        colors=tagcolors + ["k"],
        title="gf (annotation) types",
    )

    rep.len_plot(
        axes[5],
        res.match_len_by_utag,
        labels=utaglabels + ["all"],
        colors=utagcolors + ["k"],
        title="gf (annotation) types (uniq only)",
    )

    # Third row: cumulative histograms of no. matching bases
    rep.len_plot(
        axes[6],
        res.read_lengths_by_mapstat,
        title="alignment status",
        labels=alnlabels + ["all"],
        colors=alncolors + ["k"],
        xlabel="read length [nt]",
        cumulative=True,
        ylabel="cumulative fraction",
        legend=False,
    )

    # rep.len_plot(
    #     axes[8],
    #     res.match_len_by_cigtype,
    #     labels=ciglabels + ["all"],
    #     colors=cigcolors + ["k"],
    #     title="CIGAR types",
    #     cumulative=True,
    #     ylabel="cumulative fraction",
    #     legend=False,
    # )

    rep.len_plot(
        axes[7],
        res.match_len_by_tag,
        labels=taglabels + ["all"],
        colors=tagcolors + ["k"],
        title="gf (annotation) types",
        cumulative=True,
        ylabel="cumulative fraction",
        legend=False,
    )

    rep.len_plot(
        axes[8],
        res.match_len_by_utag,
        labels=utaglabels + ["all"],
        colors=utagcolors + ["k"],
        title="gf (annotation) types (uniq only)",
        cumulative=True,
        ylabel="cumulative fraction",
        legend=False,
    )

    # rep.len_plot(
    #     axes[10],
    #     res.sc_lengths,
    #     title="read position",
    #     colors=["r", "b"],
    #     xlabel="soft-clipped bases",
    #     cumulative=True,
    #     ylabel="cumulative fraction",
    #     legend=False,
    # )

    # rep.len_plot(
    #     axes[4],
    #     res.match_len_by_cigtype,
    #     labels=ciglabels + ["all"],
    #     colors=cigcolors + ["k"],
    #     title="CIGAR types",
    # )

    # rep.len_plot(
    #     axes[6],
    #     res.sc_lengths,
    #     title="read position",
    #     xlabel="soft-clipped bases",
    #     colors=["r", "b"],
    # )

    # axes[3].axis("off")

    return fig, (aln_types_d, tag_types_d, utag_types_d, mean_lens_d)  # cig_types_d,


def oligo_analysis(res):
    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    from scipy.stats import fisher_exact

    all_oligos = sorted(res.all_oligos)
    N = res.N_total
    tests = []
    interactions = []
    int_pvals = []
    int_odds = []
    table_data = []
    for oli in all_oligos:
        nuniq_w_oli = res.oligo_by_mapstat["unique"][oli]
        nuniq_wo_oli = res.aln_types["unique"] - nuniq_w_oli

        nmulti_w_oli = res.oligo_by_mapstat["multimapper"][oli]
        nmulti_wo_oli = res.aln_types["multimapper"] - nmulti_w_oli

        nmap_w_oli = res.oligo_by_mapstat["mapped"][oli]
        nmap_wo_oli = res.aln_types["mapped"] - nmap_w_oli
        nunmap_w_oli = res.oligo_by_mapstat["unmapped"][oli]
        nunmap_wo_oli = res.aln_types["unmapped"] - nunmap_w_oli

        table_map_vs_unmapped = (
            np.array([[nmap_w_oli, nmap_wo_oli], [nunmap_w_oli, nunmap_wo_oli]]) + 1
        )  # +1 = pseudo count

        table_uniq_vs_multi = (
            np.array([[nuniq_w_oli, nuniq_wo_oli], [nmulti_w_oli, nmulti_wo_oli]]) + 1
        )  # +1 = pseudo count

        for test, table in zip(
            ["mapped_vs_unmapped", "unique_vs_multimapper"],
            [table_map_vs_unmapped, table_uniq_vs_multi],
        ):
            # print(test)
            # print(table)
            # plus 1 is pseudo count
            odds, pval = fisher_exact(table + 1)
            # pval = max(pval, 1e-100)
            if pval < 0.05:
                # print(f"{test} {oli} -> odds={odds:.3e} pval={pval:.3e}")
                tests.append(test)
                interactions.append(oli)
                int_pvals.append(pval)
                int_odds.append(odds)
                table_data.append(table.ravel())

    # fig, ax = plt.subplots(figsize=(10, 7))
    # ax.spines["right"].set_visible(False)
    # ax.spines["top"].set_visible(False)
    # x = np.array(int_pvals)
    # y = np.log2(int_odds)
    # print(y)
    # ax.semilogx(x, y, ".")
    # mx, Mx = x.min(), x.max()
    # my, My = y.min(), y.max()
    # My = max(abs(my), abs(My)) * 1.1
    # wx = Mx / mx
    # wy = 2 * My

    # for x, y, label in zip(int_pvals, np.log2(int_odds), interactions):
    #     dy = (np.random.random() - 0.5) * 0.1 * wy + 0.01 * wy
    #     ax.text(x, y + dy, str(label), horizontalalignment="center")

    # ax.axhline(0, color="k", linewidth=0.5, ls="dashed")
    # ax.set_xlabel("P-value (Fisher exact)")
    # ax.set_ylabel("log2 odds ratio (oligo present vs absent)")
    # ax.set_xlim(mx * 10 ** (-0.2 * np.log10(wx)), 1)
    # ax.set_ylim(-My, My)
    # fig.tight_layout(rect=[0.2, 0.2, 0.8, 0.8])

    if len(table_data):
        a, b, c, d = np.array(table_data).T
    else:
        a, b, c, d = [], [], [], []
    df = pd.DataFrame(
        dict(
            test=tests,
            name=interactions,
            p=int_pvals,
            odds=int_odds,
            a=a,
            b=b,
            c=c,
            d=d,
        )
    )
    return df


def cmdline():
    args = parse_args()
    path = args.fname.split("/")

    if len(path) >= 4 and path[-4].startswith("sts"):
        sample_name = path[-4]
    else:
        sample_name = os.path.basename(args.fname).rsplit(".", maxsplit=1)[0]

    logging.info(f"Starting analysis of sample_name='{sample_name}'")
    res = scan_bam(
        args.fname,
        skim=args.skim,
        intact_signature=args.intact_signature,
        parse_oligos=args.parse_oligos,
    )
    # print(res.aln_types)
    if args.parse_oligos:
        df = oligo_analysis(res)
        df["sample"] = sample_name
        # fig.suptitle(sample_name)
        # fig.savefig(f"{args.out_pdf}/{sample_name}_oligo_analysis.pdf")
        # fig.savefig(f"{args.out_png}/{sample_name}_oligo_analysis.png", dpi=300)
        df.to_csv(
            f"{args.out_csv}/{sample_name}.oligo_analysis.csv", sep="\t", index=False
        )
    import matplotlib

    matplotlib.use("Agg")

    fig, (aln_counts, tag_counts, utag_counts, mean_lens_d) = make_plots(res)
    if args.intact_signature:
        fig.suptitle(f"{sample_name} ({args.intact_signature})")
    else:
        fig.suptitle(sample_name)

    if args.intact_signature:
        suffix = f"_intact_sig"
    else:
        suffix = ""

    fig.tight_layout(rect=[0, 0.03, 1, 0.95])
    fig.savefig(f"{args.out_pdf}/{sample_name}{suffix}.pdf")
    fig.savefig(f"{args.out_png}/{sample_name}{suffix}.png", dpi=300)

    # Done with the figure. Text output.
    # TODO: store as CSV instead

    rep.count_dict_out(
        res.aln_types, "alignment types", total=res.N_total, add_up="N_reads"
    )
    rep.count_dict_out(res.cigar_types, "CIGAR types", total=res.N_total)
    rep.count_dict_out(res.tag_types, "annotation (gf) types", total=res.N_total)

    # report the most frequent soft-clipped sequences
    def most_frequent(data, total, n=20, title="top"):
        print(f"### {title}")
        for k, v in sorted(data.items(), key=lambda x: -x[1])[:n]:
            # print(k, v, total)
            print(f"{k}\t{v}\t{100.0 * v / total:.3f} %")

    most_frequent(
        res.sc_seq["5'"],
        np.array(list(res.sc_seq["5'"].values())).sum(),
        title="most frequent 5' soft-clipped sequences",
    )
    most_frequent(
        res.sc_seq["3'"],
        np.array(list(res.sc_seq["3'"].values())).sum(),
        title="most frequent 3' soft-clipped sequences",
    )

    from spacemake.longread.report import count_dict_to_df

    df = pd.concat(
        [
            # count_dict_to_df(cig_counts, kind="CIGAR_types", n_total=res.N_total),
            count_dict_to_df(aln_counts, kind="align_types", n_total=res.N_total),
            count_dict_to_df(tag_counts, kind="tag_types", n_total=res.N_total),
            count_dict_to_df(
                utag_counts, kind="uniq_tag_types", n_total=aln_counts["unique"]
            ),
        ],
        ignore_index=True,
    )
    for mapstat, meanlen in mean_lens_d.items():
        df = df.append(
            dict(
                name=mapstat,
                count=meanlen,
                fraction=1.0,
                kind="mean_readlen",
            ),
            ignore_index=True,
        )

    df.to_csv(f"{args.out_csv}/{sample_name}.csv", sep="\t")


if __name__ == "__main__":
    cmdline()
